@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2020 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2020 Laszlo Molnar
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -28,7 +28,7 @@
 #include "conf.h"
 #include "linker.h"
 
-static unsigned hex(unsigned char c) { return (c & 0xf) + (c > '9' ? 9 : 0); }
+static unsigned hex(uchar c) { return (c & 0xf) + (c > '9' ? 9 : 0); }
 
 static bool update_capacity(unsigned size, unsigned *capacity) {
     if (size < *capacity)
@@ -40,12 +40,13 @@ static bool update_capacity(unsigned size, unsigned *capacity) {
     return true;
 }
 
-static void __acc_cdecl_va internal_error(const char *format, ...) {
+static void internal_error(const char *format, ...) attribute_format(1, 2);
+static void internal_error(const char *format, ...) {
     static char buf[1024];
     va_list ap;
 
     va_start(ap, format);
-    vsnprintf(buf, sizeof(buf), format, ap);
+    upx_safe_vsnprintf(buf, sizeof(buf), format, ap);
     va_end(ap);
 
     throwInternalError(buf);
@@ -56,17 +57,19 @@ static void __acc_cdecl_va internal_error(const char *format, ...) {
 **************************************************************************/
 
 ElfLinker::Section::Section(const char *n, const void *i, unsigned s, unsigned a)
-    : name(NULL), output(NULL), size(s), offset(0), p2align(a), next(NULL) {
+    : name(nullptr), output(nullptr), size(s), offset(0), p2align(a), next(nullptr) {
     name = strdup(n);
-    assert(name != NULL);
+    assert(name != nullptr);
     input = malloc(s + 1);
-    assert(input != NULL);
-    if (s != 0)
+    assert(input != nullptr);
+    if (s != 0) {
+        assert(i != nullptr);
         memcpy(input, i, s);
+    }
     ((char *) input)[s] = 0;
 }
 
-ElfLinker::Section::~Section() {
+ElfLinker::Section::~Section() noexcept {
     free(name);
     free(input);
 }
@@ -76,13 +79,13 @@ ElfLinker::Section::~Section() {
 **************************************************************************/
 
 ElfLinker::Symbol::Symbol(const char *n, Section *s, upx_uint64_t o)
-    : name(NULL), section(s), offset(o) {
+    : name(nullptr), section(s), offset(o) {
     name = strdup(n);
-    assert(name != NULL);
-    assert(section != NULL);
+    assert(name != nullptr);
+    assert(section != nullptr);
 }
 
-ElfLinker::Symbol::~Symbol() { free(name); }
+ElfLinker::Symbol::~Symbol() noexcept { free(name); }
 
 /*************************************************************************
 // Relocation
@@ -91,20 +94,16 @@ ElfLinker::Symbol::~Symbol() { free(name); }
 ElfLinker::Relocation::Relocation(const Section *s, unsigned o, const char *t, const Symbol *v,
                                   upx_uint64_t a)
     : section(s), offset(o), type(t), value(v), add(a) {
-    assert(section != NULL);
+    assert(section != nullptr);
 }
 
 /*************************************************************************
 // ElfLinker
 **************************************************************************/
 
-ElfLinker::ElfLinker()
-    : bele(&N_BELE_RTP::le_policy), input(NULL), output(NULL), head(NULL), tail(NULL),
-      sections(NULL), symbols(NULL), relocations(NULL), nsections(0), nsections_capacity(0),
-      nsymbols(0), nsymbols_capacity(0), nrelocations(0), nrelocations_capacity(0),
-      reloc_done(false) {}
+ElfLinker::ElfLinker(const N_BELE_RTP::AbstractPolicy *b) noexcept : bele(b) {}
 
-ElfLinker::~ElfLinker() {
+ElfLinker::~ElfLinker() noexcept {
     delete[] input;
     delete[] output;
 
@@ -120,8 +119,8 @@ ElfLinker::~ElfLinker() {
     free(relocations);
 }
 
-void ElfLinker::init(const void *pdata_v, int plen) {
-    const upx_byte *pdata = (const upx_byte *) pdata_v;
+void ElfLinker::init(const void *pdata_v, int plen, unsigned pxtra) {
+    const byte *pdata = (const byte *) pdata_v;
     if (plen >= 16 && memcmp(pdata, "UPX#", 4) == 0) {
         // decompress pre-compressed stub-loader
         int method;
@@ -141,71 +140,77 @@ void ElfLinker::init(const void *pdata_v, int plen) {
         }
         assert((unsigned) plen < u_len);
         inputlen = u_len;
-        input = new upx_byte[inputlen + 1];
+        input = New(byte, inputlen + 1);
         unsigned new_len = u_len;
-        int r = upx_decompress(pdata, c_len, input, &new_len, method, NULL);
+        int r = upx_decompress(pdata, c_len, input, &new_len, method, nullptr);
         if (r == UPX_E_OUT_OF_MEMORY)
             throwOutOfMemoryException();
         if (r != UPX_E_OK || new_len != u_len)
             throwBadLoader();
     } else {
         inputlen = plen;
-        input = new upx_byte[inputlen + 1];
+        input = New(byte, inputlen + 1);
         if (inputlen)
             memcpy(input, pdata, inputlen);
     }
     input[inputlen] = 0; // NUL terminate
 
-    output = new upx_byte[inputlen ? inputlen : 0x4000];
+    output_capacity = (inputlen ? (inputlen + pxtra) : 0x4000);
+    assert(output_capacity <= (1 << 16)); // LE16 l_info.l_size
+    output = New(byte, output_capacity);
     outputlen = 0;
+    NO_printf("\nElfLinker::init %d @%p\n", output_capacity, output);
 
+    // FIXME: bad compare when either symbols or relocs are absent
     if ((int) strlen("Sections:\n"
                      "SYMBOL TABLE:\n"
                      "RELOCATION RECORDS FOR ") < inputlen) {
+        char const *const eof = (char const *) &input[inputlen];
         int pos = find(input, inputlen, "Sections:\n", 10);
         assert(pos != -1);
-        char *psections = (char *) input + pos;
+        char *const psections = (char *) input + pos;
 
-        char *psymbols = strstr(psections, "SYMBOL TABLE:\n");
-        assert(psymbols != NULL);
+        char *const psymbols = strstr(psections, "SYMBOL TABLE:\n");
+        // assert(psymbols != nullptr);
 
-        char *prelocs = strstr(psymbols, "RELOCATION RECORDS FOR ");
-        assert(prelocs != NULL);
+        char *const prelocs = strstr((psymbols ? psymbols : psections), "RELOCATION RECORDS FOR ");
+        // assert(prelocs != nullptr);
 
-        preprocessSections(psections, psymbols);
-        preprocessSymbols(psymbols, prelocs);
-        preprocessRelocations(prelocs, (char *) input + inputlen);
+        preprocessSections(psections, (psymbols ? psymbols : (prelocs ? prelocs : eof)));
+        if (psymbols)
+            preprocessSymbols(psymbols, (prelocs ? prelocs : eof));
+        if (prelocs)
+            preprocessRelocations(prelocs, eof);
         addLoader("*UND*");
     }
 }
 
-void ElfLinker::preprocessSections(char *start, char *end) {
+void ElfLinker::preprocessSections(char *start, char const *end) {
     char *nextl;
     for (nsections = 0; start < end; start = 1 + nextl) {
         nextl = strchr(start, '\n');
-        assert(nextl != NULL);
+        assert(nextl != nullptr);
         *nextl = '\0'; // a record is a line
 
         unsigned offset, size, align;
         char name[1024];
 
-        if (sscanf(start, "%*d %1023s %x %*d %*d %x 2**%d", name, &size, &offset, &align) == 4) {
+        if (sscanf(start, "%*d %1023s %x %*x %*x %x 2**%d", name, &size, &offset, &align) == 4) {
             char *n = strstr(start, name);
             n[strlen(name)] = 0;
             addSection(n, input + offset, size, align);
-
-            // printf("section %s preprocessed\n", n);
+            NO_printf("section %s preprocessed\n", n);
         }
     }
-    addSection("*ABS*", NULL, 0, 0);
-    addSection("*UND*", NULL, 0, 0);
+    addSection("*ABS*", nullptr, 0, 0);
+    addSection("*UND*", nullptr, 0, 0);
 }
 
-void ElfLinker::preprocessSymbols(char *start, char *end) {
+void ElfLinker::preprocessSymbols(char *start, char const *end) {
     char *nextl;
     for (nsymbols = 0; start < end; start = 1 + nextl) {
         nextl = strchr(start, '\n');
-        assert(nextl != NULL);
+        assert(nextl != nullptr);
         *nextl = '\0'; // a record is a line
 
         unsigned value, offset;
@@ -219,30 +224,30 @@ void ElfLinker::preprocessSymbols(char *start, char *end) {
             assert(offset == 0);
         }
 #if 0
-        else if (sscanf(start, "%x%*8c %1023s %*x %1023s",
+        else if (sscanf(start, "%x%*8c %1023s %*x %1023s", &offset, section, symbol) == 3)
 #else
         // work around broken scanf() implementations
         // http://bugs.winehq.org/show_bug.cgi?id=10401 (fixed in Wine 0.9.58)
-        else if (sscanf(start, "%x%*c%*c%*c%*c%*c%*c%*c%*c %1023s %*x %1023s",
+        else if (sscanf(start, "%x%*c%*c%*c%*c%*c%*c%*c%*c %1023s %*x %1023s", &offset, section,
+                        symbol) == 3)
 #endif
-                        &offset, section, symbol) == 3)
         {
-                            char *s = strstr(start, symbol);
-                            s[strlen(symbol)] = 0;
-                            if (strcmp(section, "*UND*") == 0)
-                                offset = 0xdeaddead;
-                            assert(strcmp(section, "*ABS*") != 0);
-                            addSymbol(s, section, offset);
-                        }
+            char *s = strstr(start, symbol);
+            s[strlen(symbol)] = 0;
+            if (strcmp(section, "*UND*") == 0)
+                offset = 0xdeaddead;
+            assert(strcmp(section, "*ABS*") != 0);
+            addSymbol(s, section, offset);
+        }
     }
 }
 
-void ElfLinker::preprocessRelocations(char *start, char *end) {
-    Section *section = NULL;
+void ElfLinker::preprocessRelocations(char *start, char const *end) {
+    Section *section = nullptr;
     char *nextl;
     for (nrelocations = 0; start < end; start = 1 + nextl) {
         nextl = strchr(start, '\n');
-        assert(nextl != NULL);
+        assert(nextl != nullptr);
         *nextl = '\0'; // a record is a line
 
         {
@@ -261,7 +266,7 @@ void ElfLinker::preprocessRelocations(char *start, char *end) {
 
             upx_uint64_t add = 0;
             char *p = strstr(symbol, "+0x");
-            if (p == NULL)
+            if (p == nullptr)
                 p = strstr(symbol, "-0x");
             if (p) {
                 char sign = *p;
@@ -275,7 +280,7 @@ void ElfLinker::preprocessRelocations(char *start, char *end) {
                 else
                     add = a;
 #else
-                char *endptr = NULL;
+                char *endptr = nullptr;
                 add = strtoull(p, &endptr, 16);
                 assert(endptr && *endptr == '\0');
 #endif
@@ -285,8 +290,8 @@ void ElfLinker::preprocessRelocations(char *start, char *end) {
 
             if (section) {
                 addRelocation(section->name, offset, t, symbol, add);
-                // printf("relocation %s %s %x %llu preprocessed\n", section->name, symbol, offset,
-                // (unsigned long long) add);
+                NO_printf("relocation %s %s %x %llu preprocessed\n", section->name, symbol, offset,
+                          (unsigned long long) add);
             }
         }
     }
@@ -298,7 +303,7 @@ ElfLinker::Section *ElfLinker::findSection(const char *name, bool fatal) const {
             return sections[ic];
     if (fatal)
         internal_error("unknown section %s\n", name);
-    return NULL;
+    return nullptr;
 }
 
 ElfLinker::Symbol *ElfLinker::findSymbol(const char *name, bool fatal) const {
@@ -307,12 +312,14 @@ ElfLinker::Symbol *ElfLinker::findSymbol(const char *name, bool fatal) const {
             return symbols[ic];
     if (fatal)
         internal_error("unknown symbol %s\n", name);
-    return NULL;
+    return nullptr;
 }
 
 ElfLinker::Section *ElfLinker::addSection(const char *sname, const void *sdata, int slen,
                                           unsigned p2align) {
-    // printf("addSection: %s len=%d align=%d\n", sname, slen, p2align);
+    NO_printf("addSection: %s len=%d align=%d\n", sname, slen, p2align);
+    if (!sdata && (!strcmp("ABS*", sname) || !strcmp("UND*", sname)))
+        return nullptr;
     if (update_capacity(nsections, &nsections_capacity))
         sections =
             static_cast<Section **>(realloc(sections, nsections_capacity * sizeof(Section *)));
@@ -320,22 +327,23 @@ ElfLinker::Section *ElfLinker::addSection(const char *sname, const void *sdata,
     assert(sname);
     assert(sname[0]);
     assert(sname[strlen(sname) - 1] != ':');
-    assert(findSection(sname, false) == NULL);
+    assert(findSection(sname, false) == nullptr);
     Section *sec = new Section(sname, sdata, slen, p2align);
+    sec->sort_id = nsections;
     sections[nsections++] = sec;
     return sec;
 }
 
 ElfLinker::Symbol *ElfLinker::addSymbol(const char *name, const char *section,
                                         upx_uint64_t offset) {
-    // printf("addSymbol: %s %s 0x%x\n", name, section, offset);
+    NO_printf("addSymbol: %s %s 0x%llx\n", name, section, offset);
     if (update_capacity(nsymbols, &nsymbols_capacity))
         symbols = static_cast<Symbol **>(realloc(symbols, nsymbols_capacity * sizeof(Symbol *)));
-    assert(symbols != NULL);
+    assert(symbols != nullptr);
     assert(name);
     assert(name[0]);
     assert(name[strlen(name) - 1] != ':');
-    assert(findSymbol(name, false) == NULL);
+    assert(findSymbol(name, false) == nullptr);
     Symbol *sym = new Symbol(name, findSection(section), offset);
     symbols[nsymbols++] = sym;
     return sym;
@@ -346,7 +354,7 @@ ElfLinker::Relocation *ElfLinker::addRelocation(const char *section, unsigned of
     if (update_capacity(nrelocations, &nrelocations_capacity))
         relocations = static_cast<Relocation **>(
             realloc(relocations, (nrelocations_capacity) * sizeof(Relocation *)));
-    assert(relocations != NULL);
+    assert(relocations != nullptr);
     Relocation *rel = new Relocation(findSection(section), off, type, findSymbol(symbol), add);
     relocations[nrelocations++] = rel;
     return rel;
@@ -356,12 +364,12 @@ ElfLinker::Relocation *ElfLinker::addRelocation(const char *section, unsigned of
 void ElfLinker::setLoaderAlignOffset(int phase)
 {
     //assert(phase & 0);
-    printf("\nFIXME: ElfLinker::setLoaderAlignOffset %d\n", phase);
+    NO_printf("\nFIXME: ElfLinker::setLoaderAlignOffset %d\n", phase);
 }
 #endif
 
 int ElfLinker::addLoader(const char *sname) {
-    assert(sname != NULL);
+    assert(sname != nullptr);
     if (!sname[0])
         return outputlen;
 
@@ -395,15 +403,17 @@ int ElfLinker::addLoader(const char *sname) {
                 assert(tail);
                 assert(tail != section);
                 // .p2align must be < 32
-                unsigned const v = ~0u << section->p2align;
-                if (unsigned const l = ~v & (0u - (unsigned) (tail->offset + tail->size))) {
+                const unsigned v = ~0u << section->p2align;
+                if (const unsigned l = ~v & (0u - (unsigned) (tail->offset + tail->size))) {
                     alignCode(l);
                     tail->size += l;
                 }
             }
+            // Section->output is not relocatable, and C++ has no realloc() anyway.
+            assert((section->size + outputlen) <= output_capacity);
             memcpy(output + outputlen, section->input, section->size);
-            section->output = output + outputlen;
-            // printf("section added: 0x%04x %3d %s\n", outputlen, section->size, section->name);
+            section->output = output + outputlen; // FIXME: INVALIDATED by realloc()
+            NO_printf("section added: 0x%04x %3d %s\n", outputlen, section->size, section->name);
             outputlen += section->size;
 
             if (head) {
@@ -420,13 +430,13 @@ int ElfLinker::addLoader(const char *sname) {
 }
 
 void ElfLinker::addLoader(const char *s, va_list ap) {
-    while (s != NULL) {
+    while (s != nullptr) {
         addLoader(s);
         s = va_arg(ap, const char *);
     }
 }
 
-void __acc_cdecl_va ElfLinker::addLoaderVA(const char *s, ...) {
+void ElfLinker::addLoaderVA(const char *s, ...) {
     va_list ap;
     va_start(ap, s);
     addLoader(s, ap);
@@ -445,7 +455,7 @@ int ElfLinker::getSectionSize(const char *sname) const {
     return section->size;
 }
 
-upx_byte *ElfLinker::getLoader(int *llen) const {
+byte *ElfLinker::getLoader(int *llen) const {
     if (llen)
         *llen = outputlen;
     return output;
@@ -458,25 +468,26 @@ void ElfLinker::relocate() {
         const Relocation *rel = relocations[ic];
         upx_uint64_t value = 0;
 
-        if (rel->section->output == NULL)
+        if (rel->section->output == nullptr)
             continue;
         if (strcmp(rel->value->section->name, "*ABS*") == 0) {
             value = rel->value->offset;
         } else if (strcmp(rel->value->section->name, "*UND*") == 0 &&
                    rel->value->offset == 0xdeaddead)
             internal_error("undefined symbol '%s' referenced\n", rel->value->name);
-        else if (rel->value->section->output == NULL)
+        else if (rel->value->section->output == nullptr)
             internal_error("can not apply reloc '%s:%x' without section '%s'\n", rel->section->name,
                            rel->offset, rel->value->section->name);
         else {
             value = rel->value->section->offset + rel->value->offset + rel->add;
         }
-        upx_byte *location = rel->section->output + rel->offset;
-        // printf("%-28s %-28s %-10s %#16llx %#16llx\n", rel->section->name, rel->value->name,
-        // rel->type, (long long) value, (long long) value - rel->section->offset - rel->offset);
-        // printf("  %llx %d %llx %d %llx : %d\n", (long long) value,
-        // (int)rel->value->section->offset,
-        // rel->value->offset, rel->offset, (long long) rel->add, *location);
+        byte *location = rel->section->output + rel->offset;
+        NO_printf("%-28s %-28s %-10s %#16llx %#16llx\n", rel->section->name, rel->value->name,
+                  rel->type, (long long) value,
+                  (long long) value - rel->section->offset - rel->offset);
+        NO_printf("  %llx %d %llx %d %llx :%d\n", (long long) value,
+                  (int) rel->value->section->offset, rel->value->offset, rel->offset,
+                  (long long) rel->add, *location);
         relocate1(rel, location, value, rel->type);
     }
 }
@@ -500,21 +511,21 @@ void ElfLinker::defineSymbol(const char *name, upx_uint64_t value) {
 
 // debugging support
 void ElfLinker::dumpSymbol(const Symbol *symbol, unsigned flags, FILE *fp) const {
-    if ((flags & 1) && symbol->section->output == NULL)
+    if ((flags & 1) && symbol->section->output == nullptr)
         return;
     char d0[16 + 1], d1[16 + 1];
-    upx_snprintf(d0, sizeof(d0), "%016llx", (upx_uint64_t) symbol->offset);
-    upx_snprintf(d1, sizeof(d1), "%016llx", (upx_uint64_t) symbol->section->offset);
+    upx_safe_snprintf(d0, sizeof(d0), "%016llx", (upx_uint64_t) symbol->offset);
+    upx_safe_snprintf(d1, sizeof(d1), "%016llx", (upx_uint64_t) symbol->section->offset);
     fprintf(fp, "%-28s 0x%-16s | %-28s 0x%-16s\n", symbol->name, d0, symbol->section->name, d1);
 }
 void ElfLinker::dumpSymbols(unsigned flags, FILE *fp) const {
-    if (fp == NULL)
+    if (fp == nullptr)
         fp = stdout;
     if ((flags & 2) == 0) {
         // default: dump symbols in used section order
         for (const Section *section = head; section; section = section->next) {
             char d1[16 + 1];
-            upx_snprintf(d1, sizeof(d1), "%016llx", (upx_uint64_t) section->offset);
+            upx_safe_snprintf(d1, sizeof(d1), "%016llx", (upx_uint64_t) section->offset);
             fprintf(fp, "%-42s%-28s 0x%-16s\n", "", section->name, d1);
             for (unsigned ic = 0; ic < nsymbols; ic++) {
                 const Symbol *symbol = symbols[ic];
@@ -531,28 +542,28 @@ void ElfLinker::dumpSymbols(unsigned flags, FILE *fp) const {
 
 upx_uint64_t ElfLinker::getSymbolOffset(const char *name) const {
     const Symbol *symbol = findSymbol(name);
-    if (symbol->section->output == NULL)
+    if (symbol->section->output == nullptr)
         return 0xdeaddead;
     return symbol->section->offset + symbol->offset;
 }
 
-void ElfLinker::alignWithByte(unsigned len, unsigned char b) {
+void ElfLinker::alignWithByte(unsigned len, byte b) {
     memset(output + outputlen, b, len);
     outputlen += len;
 }
 
-void ElfLinker::relocate1(const Relocation *rel, upx_byte *, upx_uint64_t, const char *) {
+void ElfLinker::relocate1(const Relocation *rel, byte *, upx_uint64_t, const char *) {
     internal_error("unknown relocation type '%s\n'", rel->type);
 }
 
 /*************************************************************************
 // ElfLinker arch subclasses
-// FIXME: add more displacment overflow checks
+// FIXME: add more displacement overflow checks
 // FIXME: add support for our special "ignore_reloc_overflow" section
 **************************************************************************/
 
 #if 0 // FIXME
-static void check8(const Relocation *rel, const upx_byte *location, int v, int d)
+static void check8(const Relocation *rel, const byte *location, int v, int d)
 {
     if (v < -128 || v > 127)
         internal_error("value out of range (%d) in reloc %s:%x\n",
@@ -563,7 +574,7 @@ static void check8(const Relocation *rel, const upx_byte *location, int v, int d
 }
 #endif
 
-void ElfLinkerAMD64::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerAMD64::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                const char *type) {
     if (strncmp(type, "R_X86_64_", 9))
         return super::relocate1(rel, location, value, type);
@@ -571,8 +582,12 @@ void ElfLinkerAMD64::relocate1(const Relocation *rel, upx_byte *location, upx_ui
 
     bool range_check = false;
     if (strncmp(type, "PC", 2) == 0) {
-        value -= rel->section->offset + rel->offset;
         type += 2;
+        value -= rel->section->offset + rel->offset;
+        range_check = true;
+    } else if (strncmp(type, "PLT", 3) == 0) {
+        type += 3;
+        value -= rel->section->offset + rel->offset;
         range_check = true;
     }
 
@@ -592,7 +607,7 @@ void ElfLinkerAMD64::relocate1(const Relocation *rel, upx_byte *location, upx_ui
         super::relocate1(rel, location, value, type);
 }
 
-void ElfLinkerArmBE::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerArmBE::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                const char *type) {
     if (!strcmp(type, "R_ARM_PC24") || !strcmp(type, "R_ARM_CALL") ||
         !strcmp(type, "R_ARM_JUMP24")) {
@@ -617,7 +632,7 @@ void ElfLinkerArmBE::relocate1(const Relocation *rel, upx_byte *location, upx_ui
         super::relocate1(rel, location, value, type);
 }
 
-void ElfLinkerArmLE::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerArmLE::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                const char *type) {
     if (!strcmp(type, "R_ARM_PC24") || !strcmp(type, "R_ARM_CALL") ||
         !strcmp(type, "R_ARM_JUMP24")) {
@@ -642,7 +657,7 @@ void ElfLinkerArmLE::relocate1(const Relocation *rel, upx_byte *location, upx_ui
         super::relocate1(rel, location, value, type);
 }
 
-void ElfLinkerArm64LE::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerArm64LE::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                  const char *type) {
     if (strncmp(type, "R_AARCH64_", 10))
         return super::relocate1(rel, location, value, type);
@@ -690,7 +705,7 @@ void ElfLinkerM68k::alignCode(unsigned len) {
     outputlen += len;
 }
 
-void ElfLinkerM68k::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerM68k::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                               const char *type) {
     if (strncmp(type, "R_68K_", 6))
         return super::relocate1(rel, location, value, type);
@@ -710,10 +725,10 @@ void ElfLinkerM68k::relocate1(const Relocation *rel, upx_byte *location, upx_uin
         super::relocate1(rel, location, value, type);
 }
 
-void ElfLinkerMipsBE::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerMipsBE::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                 const char *type) {
-#define MIPS_HI(a) (((a) >> 16) + (((a) &0x8000) >> 15))
-#define MIPS_LO(a) ((a) &0xffff)
+#define MIPS_HI(a)   (((a) >> 16) + (((a) &0x8000) >> 15))
+#define MIPS_LO(a)   ((a) &0xffff)
 #define MIPS_PC16(a) ((a) >> 2)
 #define MIPS_PC26(a) (((a) &0x0fffffff) >> 2)
 
@@ -737,10 +752,10 @@ void ElfLinkerMipsBE::relocate1(const Relocation *rel, upx_byte *location, upx_u
 #undef MIPS_PC26
 }
 
-void ElfLinkerMipsLE::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerMipsLE::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                 const char *type) {
-#define MIPS_HI(a) (((a) >> 16) + (((a) &0x8000) >> 15))
-#define MIPS_LO(a) ((a) &0xffff)
+#define MIPS_HI(a)   (((a) >> 16) + (((a) &0x8000) >> 15))
+#define MIPS_LO(a)   ((a) &0xffff)
 #define MIPS_PC16(a) ((a) >> 2)
 #define MIPS_PC26(a) (((a) &0x0fffffff) >> 2)
 
@@ -764,7 +779,7 @@ void ElfLinkerMipsLE::relocate1(const Relocation *rel, upx_byte *location, upx_u
 #undef MIPS_PC26
 }
 
-void ElfLinkerPpc32::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerPpc32::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                const char *type) {
     if (strncmp(type, "R_PPC_", 6))
         return super::relocate1(rel, location, value, type);
@@ -785,19 +800,19 @@ void ElfLinkerPpc32::relocate1(const Relocation *rel, upx_byte *location, upx_ui
     // Note that original (*location).displ is ignored.
     if (strcmp(type, "24") == 0) {
         if (3 & value)
-            internal_error("unaligned word diplacement");
-        // FIXME: displacment overflow?
+            internal_error("unaligned word displacement");
+        // FIXME: displacement overflow?
         set_be32(location, (0xfc000003 & get_be32(location)) + (0x03fffffc & value));
     } else if (strcmp(type, "14") == 0) {
         if (3 & value)
-            internal_error("unaligned word diplacement");
-        // FIXME: displacment overflow?
+            internal_error("unaligned word displacement");
+        // FIXME: displacement overflow?
         set_be32(location, (0xffff0003 & get_be32(location)) + (0x0000fffc & value));
     } else
         super::relocate1(rel, location, value, type);
 }
 
-void ElfLinkerPpc64le::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerPpc64le::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                  const char *type) {
     if (!strcmp(type, "R_PPC64_ADDR64")) {
         set_le64(location, get_le64(location) + value);
@@ -840,7 +855,7 @@ void ElfLinkerPpc64le::relocate1(const Relocation *rel, upx_byte *location, upx_
         super::relocate1(rel, location, value, type);
 }
 
-void ElfLinkerPpc64::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerPpc64::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                                const char *type) {
     if (!strcmp(type, "R_PPC64_ADDR32")) {
         set_be32(location, get_be32(location) + value);
@@ -870,7 +885,7 @@ void ElfLinkerPpc64::relocate1(const Relocation *rel, upx_byte *location, upx_ui
         super::relocate1(rel, location, value, type);
 }
 
-void ElfLinkerX86::relocate1(const Relocation *rel, upx_byte *location, upx_uint64_t value,
+void ElfLinkerX86::relocate1(const Relocation *rel, byte *location, upx_uint64_t value,
                              const char *type) {
     if (strncmp(type, "R_386_", 6))
         return super::relocate1(rel, location, value, type);
@@ -886,7 +901,7 @@ void ElfLinkerX86::relocate1(const Relocation *rel, upx_byte *location, upx_uint
     if (strcmp(type, "8") == 0) {
         int displ = (signed char) *location + (int) value;
         if (range_check && (displ < -128 || displ > 127))
-            internal_error("target out of range (%d,%d,%d) in reloc %s:%x\n", displ, *location,
+            internal_error("target out of range (%d,%d,%llu) in reloc %s:%x\n", displ, *location,
                            value, rel->section->name, rel->offset);
         *location += value;
     } else if (strcmp(type, "16") == 0)
