@@ -1,9 +1,9 @@
-/* p_com.h --
+/* p_com.h -- dos/com executable format
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2018 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2018 Laszlo Molnar
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -25,45 +25,37 @@
    <markus@oberhumer.com>               <ezerotven+github@gmail.com>
  */
 
-
-#ifndef __UPX_P_COM_H
-#define __UPX_P_COM_H 1
-
+#pragma once
 
 /*************************************************************************
 // dos/com
 **************************************************************************/
 
-class PackCom : public Packer
-{
+class PackCom : public Packer {
     typedef Packer super;
+
 public:
-    PackCom(InputFile *f) : super(f) { bele = &N_BELE_RTP::le_policy; }
-    virtual int getVersion() const { return 13; }
-    virtual int getFormat() const { return UPX_F_DOS_COM; }
-    virtual const char *getName() const { return "dos/com"; }
-    //virtual const char *getFullName(const options_t *o) const { return o && o->cpu == o->CPU_8086 ? "i086-dos16.com" : "i286-dos16.com"; }
-    virtual const char *getFullName(const options_t *) const { return "i086-dos16.com"; }
-    virtual const int *getCompressionMethods(int method, int level) const;
-    virtual const int *getFilters() const;
+    explicit PackCom(InputFile *f) : super(f) { bele = &N_BELE_RTP::le_policy; }
+    virtual int getVersion() const override { return 13; }
+    virtual int getFormat() const override { return UPX_F_DOS_COM; }
+    virtual const char *getName() const override { return "dos/com"; }
+    virtual const char *getFullName(const Options *) const override { return "i086-dos16.com"; }
+    virtual const int *getCompressionMethods(int method, int level) const override;
+    virtual const int *getFilters() const override;
 
-    virtual void pack(OutputFile *fo);
-    virtual void unpack(OutputFile *fo);
+    virtual void pack(OutputFile *fo) override;
+    virtual void unpack(OutputFile *fo) override;
 
-    virtual bool canPack();
-    virtual int canUnpack();
+    virtual tribool canPack() override;
+    virtual tribool canUnpack() override;
 
 protected:
+    virtual Linker *newLinker() const override;
+    void addFilter16(int filter_id);
+    // dos/sys will override these:
     virtual unsigned getCallTrickOffset() const { return 0x100; }
-    virtual Linker* newLinker() const;
-
-protected:
-    virtual void buildLoader(const Filter *ft);
-    virtual void patchLoader(OutputFile *fo, upx_byte *, int, unsigned);
-    virtual void addFilter16(int filter_id);
+    virtual void buildLoader(const Filter *ft) override;
+    virtual void patchLoader(OutputFile *fo, byte *, int, unsigned);
 };
 
-
-#endif /* already included */
-
 /* vim:set ts=4 sw=4 et: */
