@@ -1,9 +1,9 @@
-/* p_exe.h --
+/* p_exe.h -- dos/exe executable format
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2020 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2020 Laszlo Molnar
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -25,56 +25,50 @@
    <markus@oberhumer.com>               <ezerotven+github@gmail.com>
  */
 
-
-#ifndef __UPX_P_EXE_H
-#define __UPX_P_EXE_H 1
-
+#pragma once
 
 /*************************************************************************
 // dos/exe
 **************************************************************************/
 
-class PackExe : public Packer
-{
+class PackExe final : public Packer {
     typedef Packer super;
+
 public:
-    PackExe(InputFile *f);
-    virtual int getVersion() const { return 13; }
-    virtual int getFormat() const { return UPX_F_DOS_EXE; }
-    virtual const char *getName() const { return "dos/exe"; }
-    //virtual const char *getFullName(const options_t *o) const { return o && o->cpu == o->CPU_8086 ? "i086-dos16.exe" : "i286-dos16.exe"; }
-    virtual const char *getFullName(const options_t *) const { return "i086-dos16.exe"; }
-    virtual const int *getCompressionMethods(int method, int level) const;
-    virtual const int *getFilters() const;
+    explicit PackExe(InputFile *f);
+    virtual int getVersion() const override { return 13; }
+    virtual int getFormat() const override { return UPX_F_DOS_EXE; }
+    virtual const char *getName() const override { return "dos/exe"; }
+    virtual const char *getFullName(const Options *) const override { return "i086-dos16.exe"; }
+    virtual const int *getCompressionMethods(int method, int level) const override;
+    virtual const int *getFilters() const override;
 
-    virtual void pack(OutputFile *fo);
-    virtual void unpack(OutputFile *fo);
+    virtual void pack(OutputFile *fo) override;
+    virtual void unpack(OutputFile *fo) override;
 
-    virtual bool canPack();
-    virtual int canUnpack();
+    virtual tribool canPack() override;
+    virtual tribool canUnpack() override;
 
     // unpacker capabilities
-    virtual bool canUnpackVersion(int version) const
-    {
+    virtual bool canUnpackVersion(int version) const override {
         // NOTE: could adapt p_exe.cpp to support (version >= 8)
         return (version >= 10);
     }
-    virtual bool canUnpackFormat(int format) const
-    {
+    virtual bool canUnpackFormat(int format) const override {
         return (format == UPX_F_DOS_EXE || format == UPX_F_DOS_EXEH);
     }
 
 protected:
     struct exe_header_t;
 
-    virtual int readFileHeader(void);
+    int readFileHeader(void);
 
-    virtual int fillExeHeader(struct exe_header_t *) const;
-    virtual void buildLoader(const Filter *ft);
-    virtual Linker* newLinker() const;
+    int fillExeHeader(struct exe_header_t *) const;
+    virtual void buildLoader(const Filter *ft) override;
+    virtual Linker *newLinker() const override;
     void addLoaderEpilogue(int flag);
 
-    __packed_struct(exe_header_t)
+    struct alignas(1) exe_header_t {
         LE16 ident;
         LE16 m512;
         LE16 p512;
@@ -84,38 +78,28 @@ protected:
         LE16 max;
         LE16 ss;
         LE16 sp;
-        char _[2];              // checksum
+        byte _[2]; // checksum
         LE16 ip;
         LE16 cs;
         LE16 relocoffs;
-        char __[2];             // overlnum
+        byte __[2]; // overlnum
         LE32 firstreloc;
-    __packed_struct_end()
+    };
 
     exe_header_t ih, oh;
 
-    unsigned ih_exesize;
-    unsigned ih_imagesize;
-    unsigned ih_overlay;
-    unsigned relocsize;
+    unsigned ih_exesize = 0;
+    unsigned ih_imagesize = 0;
+    unsigned ih_overlay = 0;
+    unsigned relocsize = 0;
 
-    bool has_9a;
-    bool device_driver;
+    bool has_9a = false;
+    bool device_driver = false;
 
-    enum {
-        NORELOC = 1,
-        USEJUMP = 2,
-        SS = 4,
-        SP = 8,
-        MINMEM = 16,
-        MAXMEM = 32
-    };
+    enum { NORELOC = 1, USEJUMP = 2, SS = 4, SP = 8, MINMEM = 16, MAXMEM = 32 };
 
-    unsigned stack_for_lzma;    // stack size required for lzma
-    bool use_clear_dirty_stack;
+    unsigned stack_for_lzma = 0; // stack size required for lzma
+    bool use_clear_dirty_stack = false;
 };
 
-
-#endif /* already included */
-
 /* vim:set ts=4 sw=4 et: */
