@@ -2,9 +2,9 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2020 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2020 Laszlo Molnar
-   Copyright (C) 2000-2020 John F. Reiser
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
+   Copyright (C) 2000-2024 John F. Reiser
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -47,23 +47,32 @@
 **************************************************************************/
 
 PackUnix::PackUnix(InputFile *f) :
-    super(f), exetype(0), blocksize(0), overlay_offset(0), lsize(0)
+    super(f), exetype(0), blocksize(0), overlay_offset(0), lsize(0),
+    methods_used(0), szb_info(sizeof(b_info))
 {
-    COMPILE_TIME_ASSERT(sizeof(Elf32_Ehdr) == 52);
-    COMPILE_TIME_ASSERT(sizeof(Elf32_Phdr) == 32);
-    COMPILE_TIME_ASSERT(sizeof(b_info) == 12);
-    COMPILE_TIME_ASSERT(sizeof(l_info) == 12);
-    COMPILE_TIME_ASSERT(sizeof(p_info) == 12);
+    COMPILE_TIME_ASSERT(sizeof(Elf32_Ehdr) == 52)
+    COMPILE_TIME_ASSERT(sizeof(Elf32_Phdr) == 32)
+    COMPILE_TIME_ASSERT(sizeof(b_info) == 12)
+    COMPILE_TIME_ASSERT(sizeof(l_info) == 12)
+    COMPILE_TIME_ASSERT(sizeof(p_info) == 12)
+
+    // Disable --android-shlib, file-by-file; undecided how to fix.
+    saved_opt_android_shlib = opt->o_unix.android_shlib;
+    opt->o_unix.android_shlib = 0;
 }
 
+PackUnix::~PackUnix()
+{
+    opt->o_unix.android_shlib = saved_opt_android_shlib;
+}
 
 // common part of canPack(), enhanced by subclasses
-bool PackUnix::canPack()
+tribool PackUnix::canPack()
 {
     if (exetype == 0)
         return false;
 
-#if defined(__unix__) && !defined(__MSYS2__)
+#if defined(__unix__)
     // must be executable by owner
     if ((fi->st.st_mode & S_IXUSR) == 0)
         throwCantPack("file not executable; try 'chmod +x'");
@@ -93,9 +102,9 @@ void PackUnix::writePackHeader(OutputFile *fo)
     set_le32(buf+0, UPX_MAGIC_LE32);
     set_le32(buf+4, UPX_MAGIC2_LE32);
 
-    checkPatch(NULL, 0, 0, 0);  // reset
+    checkPatch(nullptr, 0, 0, 0);  // reset
     patchPackHeader(buf, hsize);
-    checkPatch(NULL, 0, 0, 0);  // reset
+    checkPatch(nullptr, 0, 0, 0);  // reset
 
     fo->write(buf, hsize);
 }
@@ -137,8 +146,8 @@ int PackUnix::getStrategy(Filter &/*ft*/)
 int PackUnix::pack2(OutputFile *fo, Filter &ft)
 {
     // compress blocks
-    unsigned total_in = 0;
-    unsigned total_out = 0;
+    total_in = 0;
+    total_out = 0;
 
 // FIXME: ui_total_passes is not correct with multiple blocks...
 //    ui_total_passes = (file_size + blocksize - 1) / blocksize;
@@ -179,7 +188,7 @@ int PackUnix::pack2(OutputFile *fo, Filter &ft)
             !!n_block++);  // check compression ratio only on first block
 
         if (ph.c_len < ph.u_len) {
-            const upx_bytep tbuf = NULL;
+            const upx_bytep tbuf = nullptr;
             if (ft.id == 0) tbuf = ibuf;
             ph.overlap_overhead = OVERHEAD;
             if (!testOverlappingDecompression(obuf, tbuf, ph.overlap_overhead)) {
@@ -248,7 +257,7 @@ PackUnix::patchLoaderChecksum()
 
 off_t PackUnix::pack3(OutputFile *fo, Filter &ft)
 {
-    if (0==linker) {
+    if (nullptr==linker) {
         // If no filter, then linker is not constructed by side effect
         // of packExtent calling compressWithFilters.
         // This is typical after "/usr/bin/patchelf --set-rpath".
@@ -292,11 +301,14 @@ void PackUnix::pack(OutputFile *fo)
     fi->seek(0, SEEK_SET);
     pack1(fo, ft);  // generate Elf header, etc.
 
-    p_info hbuf;
-    set_te32(&hbuf.p_progid, progid);
-    set_te32(&hbuf.p_filesize, file_size);
-    set_te32(&hbuf.p_blocksize, blocksize);
-    fo->write(&hbuf, sizeof(hbuf));
+    // Shlib probably did not generate Elf header yet.
+    if (fo->st_size()) { // Only append if pack1 actually wrote something.
+        p_info hbuf;
+        set_te32(&hbuf.p_progid, progid);
+        set_te32(&hbuf.p_filesize, file_size);
+        set_te32(&hbuf.p_blocksize, blocksize);
+        fo->write(&hbuf, sizeof(hbuf));
+    }
 
     // append the compressed body
     if (pack2(fo, ft)) {
@@ -318,11 +330,11 @@ void PackUnix::pack(OutputFile *fo)
 
 void PackUnix::packExtent(
     const Extent &x,
-    unsigned &total_in,
-    unsigned &total_out,
     Filter *ft,
     OutputFile *fo,
-    unsigned hdr_u_len
+    unsigned hdr_u_len,
+    unsigned b_extra,
+    bool inhibit_compression_check
 )
 {
     unsigned const init_u_adler = ph.u_adler;
@@ -364,15 +376,15 @@ void PackUnix::packExtent(
             ft->cto = 0;
 
             compressWithFilters(ft, OVERHEAD, NULL_cconf, filter_strategy,
-                                0, 0, 0, hdr_ibuf, hdr_u_len);
+                                0, 0, 0, hdr_ibuf, hdr_u_len, inhibit_compression_check);
         }
         else {
             (void) compress(ibuf, ph.u_len, obuf);    // ignore return value
         }
 
         if (ph.c_len < ph.u_len) {
-            const upx_bytep tbuf = NULL;
-            if (ft == NULL || ft->id == 0) tbuf = ibuf;
+            const upx_bytep tbuf = nullptr;
+            if (ft == nullptr || ft->id == 0) tbuf = ibuf;
             ph.overlap_overhead = OVERHEAD;
             if (!testOverlappingDecompression(obuf, tbuf, ph.overlap_overhead)) {
                 // not in-place compressible
@@ -384,7 +396,7 @@ void PackUnix::packExtent(
             ph.c_len = ph.u_len;
             memcpy(obuf, ibuf, ph.c_len);
             // must update checksum of compressed data
-            ph.c_adler = upx_adler32(ibuf, ph.u_len, ph.saved_c_adler);
+            ph.c_adler = upx_adler32(ibuf, ph.u_len, init_c_adler);
         }
 
         // write block sizes
@@ -393,8 +405,8 @@ void PackUnix::packExtent(
             unsigned hdr_c_len = 0;
             MemBuffer hdr_obuf;
             hdr_obuf.allocForCompression(hdr_u_len);
-            int r = upx_compress(hdr_ibuf, hdr_u_len, hdr_obuf, &hdr_c_len, 0,
-                ph.method, 10, NULL, NULL);
+            int r = upx_compress(hdr_ibuf, hdr_u_len, hdr_obuf, &hdr_c_len, nullptr,
+                ph_forced_method(ph.method), 10, nullptr, nullptr);
             if (r != UPX_E_OK)
                 throwInternalError("header compression failed");
             if (hdr_c_len >= hdr_u_len)
@@ -407,8 +419,10 @@ void PackUnix::packExtent(
             memset(&tmp, 0, sizeof(tmp));
             set_te32(&tmp.sz_unc, hdr_u_len);
             set_te32(&tmp.sz_cpr, hdr_c_len);
-            tmp.b_method = (unsigned char) ph.method;
+            tmp.b_method = (unsigned char) ph_forced_method(ph.method);
+            tmp.b_extra = b_extra;
             fo->write(&tmp, sizeof(tmp));
+            total_out += sizeof(tmp);
             b_len += sizeof(b_info);
             fo->write(hdr_obuf, hdr_c_len);
             total_out += hdr_c_len;
@@ -425,7 +439,9 @@ void PackUnix::packExtent(
                 tmp.b_cto8 = ft->cto;
             }
         }
+        tmp.b_extra = b_extra;
         fo->write(&tmp, sizeof(tmp));
+        total_out += sizeof(tmp);
         b_len += sizeof(b_info);
 
         if (ft) {
@@ -434,25 +450,31 @@ void PackUnix::packExtent(
         // write compressed data
         if (ph.c_len < ph.u_len) {
             fo->write(obuf, ph.c_len);
+            total_out += ph.c_len;
             // Checks ph.u_adler after decompression, after unfiltering
             verifyOverlappingDecompression(ft);
         }
         else {
             fo->write(ibuf, ph.u_len);
+            total_out += ph.u_len;
         }
 
         total_in += ph.u_len;
-        total_out += ph.c_len;
     }
 }
 
-void PackUnix::unpackExtent(unsigned wanted, OutputFile *fo,
-    unsigned &total_in, unsigned &total_out,
+// Consumes b_info header block and sz_cpr data block from input file 'fi'.
+// De-compresses; appends to output file 'fo' unless rewrite or peeking.
+// For "peeking" without writing: set (fo = nullptr), (is_rewrite = -1)
+// Return actual length when peeking; else 0.
+unsigned PackUnix::unpackExtent(unsigned wanted, OutputFile *fo,
     unsigned &c_adler, unsigned &u_adler,
-    bool first_PF_X, unsigned szb_info, bool is_rewrite
+    bool first_PF_X,
+    int is_rewrite // 0(false): write; 1(true): rewrite; -1: no write
 )
 {
     b_info hdr; memset(&hdr, 0, sizeof(hdr));
+    unsigned inlen = 0; // output index (if-and-only-if peeking)
     while (wanted) {
         fi->readx(&hdr, szb_info);
         int const sz_unc = ph.u_len = get_te32(&hdr.sz_unc);
@@ -468,20 +490,22 @@ void PackUnix::unpackExtent(unsigned wanted, OutputFile *fo,
         if (sz_cpr > sz_unc || sz_unc > (int)blocksize)
             throwCantUnpack("corrupt b_info");
 
-        int j = blocksize + OVERHEAD - sz_cpr;
+        // place the input for overlapping de-compression
+        // FIXME: inlen cheats OVERHEAD; assumes small wanted peek length
+        int j = inlen + blocksize + OVERHEAD - sz_cpr;
         fi->readx(ibuf+j, sz_cpr);
+        total_in += sz_cpr;
         // update checksum of compressed data
         c_adler = upx_adler32(ibuf + j, sz_cpr, c_adler);
-        // decompress
-        if (sz_cpr < sz_unc)
-        {
-            decompress(ibuf+j, ibuf, false);
+
+        if (sz_cpr < sz_unc) { // block was compressed
+            decompress(ibuf+j, ibuf+inlen, false);
             if (12==szb_info) { // modern per-block filter
                 if (hdr.b_ftid) {
                     Filter ft(ph.level);  // FIXME: ph.level for b_info?
                     ft.init(hdr.b_ftid, 0);
                     ft.cto = hdr.b_cto8;
-                    ft.unfilter(ibuf, sz_unc);
+                    ft.unfilter(ibuf+inlen, sz_unc);
                 }
             }
             else { // ancient per-file filter
@@ -492,35 +516,46 @@ void PackUnix::unpackExtent(unsigned wanted, OutputFile *fo,
                     Filter ft(ph.level);
                     ft.init(ph.filter, 0);
                     ft.cto = (unsigned char) ph.filter_cto;
-                    ft.unfilter(ibuf, sz_unc);
+                    ft.unfilter(ibuf+inlen, sz_unc);
                 }
             }
-            j = 0;
+        }
+        else if (sz_cpr == sz_unc) { // slide literal (non-compressible) block
+            memmove(&ibuf[inlen], &ibuf[j], sz_unc);
         }
         // update checksum of uncompressed data
-        u_adler = upx_adler32(ibuf + j, sz_unc, u_adler);
-        total_in  += sz_cpr;
-        total_out += sz_unc;
+        u_adler = upx_adler32(ibuf + inlen, sz_unc, u_adler);
         // write block
         if (fo) {
             if (is_rewrite) {
-                fo->rewrite(ibuf + j, sz_unc);
+                fo->rewrite(ibuf, sz_unc);
             }
             else {
-                fo->write(ibuf + j, sz_unc);
+                fo->write(ibuf, sz_unc);
+                total_out += sz_unc;
             }
         }
-        if (wanted < (unsigned)sz_unc)
+        else if (is_rewrite < 0) { // append to &ibuf[inlen]
+            inlen += sz_unc;  // accounting; data is already there
+            if (wanted <= (unsigned)sz_unc)  // done
+                break;
+        }
+        else if (wanted < (unsigned)sz_unc) // mismatched end-of-block
             throwCantUnpack("corrupt b_info");
+        else { // "upx -t": (!fo && !(is_rewrite < 0))
+            // No output.
+        }
         wanted -= sz_unc;
     }
+    return inlen;
 }
 
 /*************************************************************************
 // Generic Unix canUnpack().
 **************************************************************************/
 
-int PackUnix::canUnpack()
+// The prize is the value of overlay_offset: the offset of compressed data
+tribool PackUnix::canUnpack()
 {
     int const small = 32 + sizeof(overlay_offset);
     // Allow zero-filled last page, for Mac OS X code signing.
@@ -531,6 +566,13 @@ int PackUnix::canUnpack()
 
     fi->seek(-(off_t)bufsize, SEEK_END);
     fi->readx(buf, bufsize);
+    return find_overlay_offset(buf);
+}
+
+int PackUnix::find_overlay_offset(MemBuffer const &buf)
+{
+    int const small = 32 + sizeof(overlay_offset);
+    int const bufsize = buf.getSize();
     int i = bufsize;
     while (i > small && 0 == buf[--i]) { }
     i -= small;
@@ -548,7 +590,6 @@ int PackUnix::canUnpack()
     return true;
 }
 
-
 /*************************************************************************
 // Generic Unix unpack().
 //
@@ -559,13 +600,12 @@ int PackUnix::canUnpack()
 void PackUnix::unpack(OutputFile *fo)
 {
     b_info bhdr;
-    unsigned const szb_info = (ph.version <= 11)
-        ? sizeof(bhdr.sz_unc) + sizeof(bhdr.sz_cpr)  // old style
-        : sizeof(bhdr);
-
-    unsigned c_adler = upx_adler32(NULL, 0);
-    unsigned u_adler = upx_adler32(NULL, 0);
+    unsigned c_adler = upx_adler32(nullptr, 0);
+    unsigned u_adler = upx_adler32(nullptr, 0);
 
+    if (ph.version <= 11) {
+        szb_info = sizeof(bhdr.sz_unc) + sizeof(bhdr.sz_cpr);  // old style
+    }
     // defaults for ph.version == 8
     unsigned orig_file_size = 0;
     blocksize = 512 * 1024;
@@ -592,8 +632,8 @@ void PackUnix::unpack(OutputFile *fo)
     ibuf.alloc(blocksize + OVERHEAD);
 
     // decompress blocks
-    unsigned total_in = 0;
-    unsigned total_out = 0;
+    total_in = 0;
+    total_out = 0;
     memset(&bhdr, 0, sizeof(bhdr));
     for (;;)
     {
