@@ -2,9 +2,9 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 2004-2020 John Reiser
-   Copyright (C) 1996-2020 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2020 Laszlo Molnar
+   Copyright (C) 2004-2024 John Reiser
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -38,15 +38,15 @@
 #include "p_vmlinx.h"
 #include "linker.h"
 
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/i386-linux.kernel.vmlinux.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/amd64-linux.kernel.vmlinux.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/arm.v5a-linux.kernel.vmlinux.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/armeb.v5a-linux.kernel.vmlinux.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/powerpc-linux.kernel.vmlinux.h"
 
 
@@ -61,10 +61,10 @@ PackVmlinuxBase<T>::PackVmlinuxBase(InputFile *f,
     super(f),
     my_e_machine(e_machine), my_elfclass(elfclass), my_elfdata(elfdata),
     my_boot_label(boot_label),
-    n_ptload(0), phdri(NULL), shdri(NULL), shstrtab(NULL)
+    n_ptload(0), phdri(nullptr), shdri(nullptr), shstrtab(nullptr)
 {
     ElfClass::compileTimeAssertions();
-    bele = N_BELE_CTP::getRTP((const BeLePolicy*) NULL);
+    bele = N_BELE_CTP::getRTP((const BeLePolicy*) nullptr);
 }
 
 template <class T>
@@ -98,41 +98,63 @@ PackVmlinuxBase<T>::compare_Phdr(void const *aa, void const *bb)
             if (xa > xb)         return  1;
     if (a->p_paddr < b->p_paddr) return -1;  // ascending by .p_paddr
     if (a->p_paddr > b->p_paddr) return  1;
-    return  0;
+    // What could remain?
+    // try to make sort order deterministic and just compare more fields
+#define CMP(field) \
+    if (a->field != b->field) return a->field < b->field ? -1 : 1
+    CMP(p_offset);
+    CMP(p_vaddr);
+    CMP(p_filesz);
+    CMP(p_memsz);
+    CMP(p_flags);
+    CMP(p_align);
+#undef CMP
+    return 0;
 }
 
 template <class T>
 typename T::Shdr const *PackVmlinuxBase<T>::getElfSections()
 {
-    Shdr const *p, *shstrsec=0;
-    shdri = new Shdr[(unsigned) ehdri.e_shnum];
+    unsigned const e_shnum = ehdri.e_shnum;
+    if (ehdri.e_shentsize != sizeof(*shdri)
+    ||  file_size_u < ehdri.e_shoff
+    ||  file_size_u < ehdri.e_shoff + mem_size(ehdri.e_shentsize, e_shnum)) {
+        infoWarning("bad ElfXX_Shdrs");
+        return nullptr;
+    }
+    shdri = new Shdr[(unsigned) e_shnum];
     fi->seek(ehdri.e_shoff, SEEK_SET);
-    fi->readx(shdri, ehdri.e_shnum * sizeof(*shdri));
-    int j;
-    for (p = shdri, j= ehdri.e_shnum; --j>=0; ++p) {
-        if (Shdr::SHT_STRTAB==p->sh_type
-        &&  p->sh_offset  <   (unsigned long)file_size
-        &&  p->sh_size    <= ((unsigned long)file_size - p->sh_offset)
-        &&  p->sh_name    <   (unsigned long)file_size
-        &&  10            <= ((unsigned long)file_size - p->sh_name)
-            // 10 == (1+ strlen(".shstrtab"))
-        ) {
-            delete [] shstrtab;
-            shstrtab = new char[1+ p->sh_size];
-            fi->seek(p->sh_offset, SEEK_SET);
-            fi->readx(shstrtab, p->sh_size);
-            shstrtab[p->sh_size] = '\0';
-            if (0==strcmp(".shstrtab", shstrtab + p->sh_name)) {
-                shstrsec = p;
-                break;
-            }
+    fi->readx(shdri, e_shnum * sizeof(*shdri));
+    unsigned const e_shstrndx = ehdri.e_shstrndx;
+    if (e_shnum <= e_shstrndx) {
+        infoWarning("bad .e_shstrndx %#x", e_shstrndx);
+        return nullptr;
+    }
+    Shdr const *p = &shdri[e_shstrndx];
+    if (Shdr::SHT_STRTAB==p->sh_type
+    &&  p->sh_offset  <= (file_size_u - sizeof(*shdri))
+    &&  p->sh_size    <= (file_size_u - p->sh_offset)
+    &&  p->sh_name    <= (file_size_u - p->sh_offset)
+    &&  10            <= (file_size_u - p->sh_name)
+        // 10 == (1+ strlen(".shstrtab"))
+    ) {
+        if (p->sh_size <= p->sh_name) {
+            infoWarning("bad .shstrtab _Shdr[%u]", (unsigned)ehdri.e_shstrndx);
+            return nullptr;
+        }
+        shstrtab = new char[1+ p->sh_size];
+        fi->seek(p->sh_offset, SEEK_SET);
+        fi->readx(shstrtab, p->sh_size);
+        shstrtab[p->sh_size] = '\0';
+        if (0==strcmp(".shstrtab", shstrtab + p->sh_name)) {
+            return p;
         }
     }
-    return shstrsec;
+    return nullptr;
 }
 
 template <class T>
-bool PackVmlinuxBase<T>::canPack()
+tribool PackVmlinuxBase<T>::canPack()
 {
     fi->seek(0, SEEK_SET);
     fi->readx(&ehdri, sizeof(ehdri));
@@ -161,14 +183,14 @@ bool PackVmlinuxBase<T>::canPack()
 
     // A Linux kernel must have a __ksymtab section. [??]
     Shdr const *p, *const shstrsec = getElfSections();
-    if (0==shstrsec) {
+    if (nullptr==shstrsec) {
         return false;
     }
     {
         int j;
         for (p = shdri, j= ehdri.e_shnum; --j>=0; ++p) {
             if (Shdr::SHT_PROGBITS==p->sh_type
-            && (p->sh_name  + shstrsec->sh_offset) < (unsigned)file_size
+            && p->sh_name < shstrsec->sh_size
             && 0==strcmp("__ksymtab", p->sh_name + shstrtab)) {
                 break;
             }
@@ -183,7 +205,7 @@ bool PackVmlinuxBase<T>::canPack()
     fi->readx(phdri, ehdri.e_phnum * sizeof(*phdri));
 
     // Put PT_LOAD together at the beginning, ascending by .p_paddr.
-    qsort(phdri, ehdri.e_phnum, sizeof(*phdri), compare_Phdr);
+    upx_qsort(phdri, ehdri.e_phnum, sizeof(*phdri), compare_Phdr);
 
     // Find convex hull of physical addresses, and count the PT_LOAD.
     // Ignore ".bss": .p_filesz < .p_memsz
@@ -218,7 +240,7 @@ void PackVmlinuxBase<T>::pack(OutputFile *fo)
     Ehdr ehdro;
     TE32 tmp_u32;
 
-    // NULL
+    // nullptr
     // .text(PT_LOADs) .note(1st page) .note(rest)
     // .shstrtab .symtab .strtab
     Shdr shdro[1+3+3];
@@ -263,7 +285,7 @@ void PackVmlinuxBase<T>::pack(OutputFile *fo)
             fi->readx(ibuf + ((unsigned) phdri[j].p_paddr - paddr_min), phdri[j].p_filesz);
         }
     }
-    checkAlreadyPacked(ibuf + ph.u_len - 1024, 1024);
+    checkAlreadyPacked(ibuf + (ph.u_len - 1024), 1024);
 
     // prepare filter
     ph.filter = 0;
@@ -301,19 +323,19 @@ void PackVmlinuxBase<T>::pack(OutputFile *fo)
         memcpy(&unc_hdr[sizeof(ehdri)],  phdri, sizeof(Phdr) * ehdri.e_phnum);
         unsigned len_cpr = 0;
         int const r = upx_compress(unc_hdr, len_unc, cpr_hdr, &len_cpr,
-            NULL, ph.method, 10, NULL, NULL );
+            nullptr, ph.method, 10, nullptr, nullptr );
         if (UPX_E_OK!=r || len_unc<=len_cpr)  // FIXME: allow no compression
             throwInternalError("Ehdr compression failed");
 
-        __packed_struct(b_info) // 12-byte header before each compressed block
-            unsigned sz_unc;  // uncompressed_size
-            unsigned sz_cpr;  //   compressed_size
+        packed_struct(b_info) { // 12-byte header before each compressed block
+            NE32 sz_unc;  // uncompressed_size
+            NE32 sz_cpr;  //   compressed_size
             unsigned char b_method;  // compression algorithm
             unsigned char b_ftid;  // filter id
             unsigned char b_cto8;  // filter parameter
             unsigned char b_unused;  // FIXME: !=0  for partial-block unfilter
-            // unsigned f_offset, f_len;  // only if    partial-block unfilter
-        __packed_struct_end()
+            // NE32 f_offset, f_len;  // only if    partial-block unfilter
+        };
 
         struct b_info hdr_info;
         set_be32(&hdr_info.sz_unc, len_unc);
@@ -359,7 +381,7 @@ void PackVmlinuxBase<T>::pack(OutputFile *fo)
         }
         compressWithFilters(ibuf, ph.u_len, obuf,
             f_ptr, f_len,  // filter range
-            NULL, 0,  // hdr_ptr, hdr_len
+            nullptr, 0,  // hdr_ptr, hdr_len
             &ft, 512, &cconf, getStrategy(ft));
 
         set_be32(&hdr_info.sz_unc, ph.u_len);
@@ -542,7 +564,7 @@ void PackVmlinuxBase<T>::pack(OutputFile *fo)
 }
 
 template <class T>
-int PackVmlinuxBase<T>::canUnpack()
+tribool PackVmlinuxBase<T>::canUnpack()
 {
     fi->seek(0, SEEK_SET);
     fi->readx(&ehdri, sizeof(ehdri));
@@ -569,12 +591,12 @@ int PackVmlinuxBase<T>::canUnpack()
 
     // find the .shstrtab section
     Shdr const *const shstrsec = getElfSections();
-    if (0==shstrsec) {
+    if (nullptr==shstrsec) {
         return false;
     }
 
     // check for .text .note .note  and sane (.sh_size + .sh_offset)
-    p_note0 = p_note1 = p_text = 0;
+    p_note0 = p_note1 = p_text = nullptr;
     int j;
     Shdr *p;
     for (p= shdri, j= ehdri.e_shnum; --j>=0; ++p) {
@@ -587,15 +609,15 @@ int PackVmlinuxBase<T>::canUnpack()
             p_text = p;
         }
         if (0==strcmp(".note", shstrtab + p->sh_name)) {
-            if (0==p_note0) {
+            if (nullptr==p_note0) {
                 p_note0 = p;
             } else
-            if (0==p_note1) {
+            if (nullptr==p_note1) {
                 p_note1 = p;
             }
         }
     }
-    if (0==p_text || 0==p_note0 || 0==p_note1) {
+    if (nullptr==p_text || nullptr==p_note0 || nullptr==p_note1) {
         return false;
     }
 
@@ -620,7 +642,7 @@ void PackVmlinuxBase<T>::unpack(OutputFile *fo)
     ph.c_len = p_note0->sh_size - sizeof(word);
     ibuf.alloc(ph.c_len);
     fi->readx(ibuf, ph.c_len);
-    obuf.allocForUncompression(ph.u_len);
+    obuf.allocForDecompression(ph.u_len);
     decompress(ibuf, obuf, false);
     fo->write(obuf, ph.u_len);
     obuf.dealloc();
@@ -632,7 +654,7 @@ void PackVmlinuxBase<T>::unpack(OutputFile *fo)
     }
     ibuf.alloc(ph.c_len);
     fi->readx(ibuf, ph.c_len);
-    obuf.allocForUncompression(ph.u_len);
+    obuf.allocForDecompression(ph.u_len);
     decompress(ibuf, obuf);
 
     Filter ft(ph.level);
@@ -649,7 +671,7 @@ void PackVmlinuxBase<T>::unpack(OutputFile *fo)
     ph.c_len = p_note1->sh_size - sizeof(word);
     ibuf.alloc(ph.c_len);
     fi->readx(ibuf, p_note1->sh_size - sizeof(ph.u_len));
-    obuf.allocForUncompression(ph.u_len);
+    obuf.allocForDecompression(ph.u_len);
     decompress(ibuf, obuf, false);
     fo->write(obuf, ph.u_len);
     obuf.dealloc();
@@ -769,21 +791,21 @@ void PackVmlinuxI386::buildLoader(const Filter *ft)
               (0x40==(0xf0 & ft->id)) ? "LXCKLLT1" : (ft->id ? "LXCALLT1" : ""),
               "LXMOVEUP",
               getDecompressorSections(),
-              NULL
+              nullptr
              );
     if (ft->id) {
         assert(ft->calls > 0);
         if (0x40==(0xf0 & ft->id)) {
-            addLoader("LXCKLLT9", NULL);
+            addLoader("LXCKLLT9", nullptr);
         }
         else {
-            addLoader("LXCALLT9", NULL);
+            addLoader("LXCALLT9", nullptr);
         }
         addFilter32(ft->id);
     }
     addLoader("LINUX990",
               ((ph.first_offset_found == 1) ? "LINUX991" : ""),
-              "LINUX992,IDENTSTR,UPX1HEAD", NULL);
+              "LINUX992,IDENTSTR,UPX1HEAD", nullptr);
 }
 
 void PackVmlinuxAMD64::buildLoader(const Filter *ft)
@@ -794,21 +816,21 @@ void PackVmlinuxAMD64::buildLoader(const Filter *ft)
               (0x40==(0xf0 & ft->id)) ? "LXCKLLT1" : (ft->id ? "LXCALLT1" : ""),
               "LXMOVEUP",
               getDecompressorSections(),
-              NULL
+              nullptr
              );
     if (ft->id) {
         assert(ft->calls > 0);
         if (0x40==(0xf0 & ft->id)) {
-            addLoader("LXCKLLT9", NULL);
+            addLoader("LXCKLLT9", nullptr);
         }
         else {
-            addLoader("LXCALLT9", NULL);
+            addLoader("LXCALLT9", nullptr);
         }
         addFilter32(ft->id);
     }
     addLoader("LINUX990",
               ((ph.first_offset_found == 1) ? "LINUX991" : ""),
-              "LINUX992,IDENTSTR,UPX1HEAD", NULL);
+              "LINUX992,IDENTSTR,UPX1HEAD", nullptr);
 }
 
 bool PackVmlinuxARMEL::is_valid_e_entry(Addr e_entry)
@@ -856,108 +878,108 @@ void PackVmlinuxARMEL::buildLoader(const Filter *ft)
 {
     // prepare loader
     initLoader(stub_arm_v5a_linux_kernel_vmlinux, sizeof(stub_arm_v5a_linux_kernel_vmlinux));
-    addLoader("LINUX000", NULL);
+    addLoader("LINUX000", nullptr);
     if (ft->id) {
         assert(ft->calls > 0);
-        addLoader("LINUX010", NULL);
+        addLoader("LINUX010", nullptr);
     }
-    addLoader("LINUX020", NULL);
+    addLoader("LINUX020", nullptr);
     if (ft->id) {
         addFilter32(ft->id);
     }
-    addLoader("LINUX030", NULL);
-         if (ph.method == M_NRV2E_8) addLoader("NRV2E", NULL);
-    else if (ph.method == M_NRV2B_8) addLoader("NRV2B", NULL);
-    else if (ph.method == M_NRV2D_8) addLoader("NRV2D", NULL);
-    else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00,LZMA_DEC10,LZMA_DEC30", NULL);
+    addLoader("LINUX030", nullptr);
+         if (ph.method == M_NRV2E_8) addLoader("NRV2E", nullptr);
+    else if (ph.method == M_NRV2B_8) addLoader("NRV2B", nullptr);
+    else if (ph.method == M_NRV2D_8) addLoader("NRV2D", nullptr);
+    else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00,LZMA_DEC10,LZMA_DEC30", nullptr);
     else throwBadLoader();
-    addLoader("IDENTSTR,UPX1HEAD", NULL);
+    addLoader("IDENTSTR,UPX1HEAD", nullptr);
 }
 
 void PackVmlinuxARMEB::buildLoader(const Filter *ft)
 {
     // prepare loader
     initLoader(stub_armeb_v5a_linux_kernel_vmlinux, sizeof(stub_armeb_v5a_linux_kernel_vmlinux));
-    addLoader("LINUX000", NULL);
+    addLoader("LINUX000", nullptr);
     if (ft->id) {
         assert(ft->calls > 0);
-        addLoader("LINUX010", NULL);
+        addLoader("LINUX010", nullptr);
     }
-    addLoader("LINUX020", NULL);
+    addLoader("LINUX020", nullptr);
     if (ft->id) {
         addFilter32(ft->id);
     }
-    addLoader("LINUX030", NULL);
-         if (ph.method == M_NRV2E_8) addLoader("NRV2E", NULL);
-    else if (ph.method == M_NRV2B_8) addLoader("NRV2B", NULL);
-    else if (ph.method == M_NRV2D_8) addLoader("NRV2D", NULL);
-    else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00,LZMA_DEC10,LZMA_DEC30", NULL);
+    addLoader("LINUX030", nullptr);
+         if (ph.method == M_NRV2E_8) addLoader("NRV2E", nullptr);
+    else if (ph.method == M_NRV2B_8) addLoader("NRV2B", nullptr);
+    else if (ph.method == M_NRV2D_8) addLoader("NRV2D", nullptr);
+    else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00,LZMA_DEC10,LZMA_DEC30", nullptr);
     else throwBadLoader();
-    addLoader("IDENTSTR,UPX1HEAD", NULL);
+    addLoader("IDENTSTR,UPX1HEAD", nullptr);
 }
 
 void PackVmlinuxPPC32::buildLoader(const Filter *ft)
 {
     // prepare loader
     initLoader(stub_powerpc_linux_kernel_vmlinux, sizeof(stub_powerpc_linux_kernel_vmlinux));
-    addLoader("LINUX000", NULL);
+    addLoader("LINUX000", nullptr);
     if (ft->id) {
         assert(ft->calls > 0);
-        addLoader("LINUX010", NULL);
+        addLoader("LINUX010", nullptr);
     }
-    addLoader("LINUX020", NULL);
+    addLoader("LINUX020", nullptr);
     if (ft->id) {
         addFilter32(ft->id);
     }
-    addLoader("LINUX030", NULL);
-         if (ph.method == M_NRV2E_LE32) addLoader("NRV2E,NRV_TAIL", NULL);
-    else if (ph.method == M_NRV2B_LE32) addLoader("NRV2B,NRV_TAIL", NULL);
-    else if (ph.method == M_NRV2D_LE32) addLoader("NRV2D,NRV_TAIL", NULL);
-    else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00,LZMA_DEC10,LZMA_DEC30", NULL);
+    addLoader("LINUX030", nullptr);
+         if (ph.method == M_NRV2E_LE32) addLoader("NRV2E,NRV_TAIL", nullptr);
+    else if (ph.method == M_NRV2B_LE32) addLoader("NRV2B,NRV_TAIL", nullptr);
+    else if (ph.method == M_NRV2D_LE32) addLoader("NRV2D,NRV_TAIL", nullptr);
+    else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00,LZMA_DEC10,LZMA_DEC30", nullptr);
     else throwBadLoader();
     if (hasLoaderSection("CFLUSH"))
         addLoader("CFLUSH");
 
-    addLoader("IDENTSTR,UPX1HEAD", NULL);
+    addLoader("IDENTSTR,UPX1HEAD", nullptr);
 }
 
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/powerpc64le-linux.kernel.vmlinux.h"
 void PackVmlinuxPPC64LE::buildLoader(const Filter *ft)
 {
     // prepare loader
     initLoader(stub_powerpc64le_linux_kernel_vmlinux, sizeof(stub_powerpc64le_linux_kernel_vmlinux));
-    addLoader("LINUX000", NULL);
+    addLoader("LINUX000", nullptr);
     if (ft->id) {
         assert(ft->calls > 0);
-        addLoader("LINUX010", NULL);
+        addLoader("LINUX010", nullptr);
     }
-    addLoader("LINUX020", NULL);
+    addLoader("LINUX020", nullptr);
     if (ft->id) {
         addFilter32(ft->id);
     }
-    addLoader("LINUX030", NULL);
-         if (ph.method == M_NRV2E_LE32) addLoader("NRV2E,NRV_TAIL", NULL);
-    else if (ph.method == M_NRV2B_LE32) addLoader("NRV2B,NRV_TAIL", NULL);
-    else if (ph.method == M_NRV2D_LE32) addLoader("NRV2D,NRV_TAIL", NULL);
-    else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00,LZMA_DEC10,LZMA_DEC30", NULL);
+    addLoader("LINUX030", nullptr);
+         if (ph.method == M_NRV2E_LE32) addLoader("NRV2E,NRV_TAIL", nullptr);
+    else if (ph.method == M_NRV2B_LE32) addLoader("NRV2B,NRV_TAIL", nullptr);
+    else if (ph.method == M_NRV2D_LE32) addLoader("NRV2D,NRV_TAIL", nullptr);
+    else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00,LZMA_DEC10,LZMA_DEC30", nullptr);
     else throwBadLoader();
     if (hasLoaderSection("CFLUSH"))
         addLoader("CFLUSH");
 
-    addLoader("IDENTSTR,UPX1HEAD", NULL);
+    addLoader("IDENTSTR,UPX1HEAD", nullptr);
 }
 
 
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/i386-linux.kernel.vmlinux-head.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/amd64-linux.kernel.vmlinux-head.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/arm.v5a-linux.kernel.vmlinux-head.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/armeb.v5a-linux.kernel.vmlinux-head.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/powerpc-linux.kernel.vmlinux-head.h"
 
 unsigned PackVmlinuxI386::write_vmlinux_head(
@@ -1144,6 +1166,7 @@ bool PackVmlinuxPPC32::has_valid_vmlinux_head()
     return false;
 }
 
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/powerpc64le-linux.kernel.vmlinux-head.h"
 bool PackVmlinuxPPC64LE::has_valid_vmlinux_head()
 {
@@ -1215,8 +1238,9 @@ Linker* PackVmlinuxAMD64::newLinker() const
 
 
 // instantiate instances
-template class PackVmlinuxBase<ElfClass_LE32>;
 template class PackVmlinuxBase<ElfClass_BE32>;
+// template class PackVmlinuxBase<ElfClass_BE64>; // not used
+template class PackVmlinuxBase<ElfClass_LE32>;
 template class PackVmlinuxBase<ElfClass_LE64>;
 
 /* vim:set ts=4 sw=4 et: */
