@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2020 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2020 Laszlo Molnar
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -26,6 +26,7 @@
  */
 
 
+#define ALLOW_INT_PLUS_MEMBUFFER 1
 #include "conf.h"
 
 #include "p_elf.h"
@@ -34,9 +35,9 @@
 #include "packer.h"
 #include "p_vmlinz.h"
 #include "linker.h"
-#include <zlib.h>
+#include <zlib/zlib.h>
 
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/i386-linux.kernel.vmlinuz.h"
 
 static const unsigned stack_offset_during_uncompression = 0x9000;
@@ -56,7 +57,7 @@ PackVmlinuzI386::PackVmlinuzI386(InputFile *f) :
     , filter_len(0)
 {
     bele = &N_BELE_RTP::le_policy;
-    COMPILE_TIME_ASSERT(sizeof(boot_sect_t) == 0x250);
+    COMPILE_TIME_ASSERT(sizeof(boot_sect_t) == 0x250)
 }
 
 
@@ -92,7 +93,7 @@ int PackVmlinuzI386::getStrategy(Filter &/*ft*/)
 }
 
 
-bool PackVmlinuzI386::canPack()
+tribool PackVmlinuzI386::canPack()
 {
     return readFileHeader() == getFormat();
 }
@@ -116,7 +117,7 @@ int PackVmlinuzI386::readFileHeader()
         return 0;
 
     int format = UPX_F_VMLINUZ_i386;
-    unsigned sys_size = ALIGN_UP((unsigned) file_size, 16u) - setup_size;
+    unsigned sys_size = ALIGN_UP((unsigned) file_size_u, 16u) - setup_size;
 
     const unsigned char *p = (const unsigned char *) &h + 0x1e3;
 
@@ -159,7 +160,7 @@ int PackVmlinuzI386::decompressKernel()
     fi->readx(obuf, file_size);
 
     {
-    const upx_byte *base = NULL;
+    const upx_byte *base = nullptr;
     unsigned relocated = 0;
 
     // See startup_32: in linux/arch/i386/boot/compressed/head.S
@@ -234,7 +235,7 @@ int PackVmlinuzI386::decompressKernel()
     }
     }
 
-    checkAlreadyPacked(obuf + setup_size, UPX_MIN(file_size - setup_size, (off_t)1024));
+    checkAlreadyPacked(obuf + setup_size, UPX_MIN(file_size - setup_size, 1024LL));
 
     int gzoff = setup_size;
     if (0x208<=h.version) {
@@ -259,7 +260,7 @@ int PackVmlinuzI386::decompressKernel()
         // try to decompress
         int klen;
         int fd;
-        off_t fd_pos;
+        upx_off_t fd_pos;
         for (;;)
         {
             klen = -1;
@@ -270,7 +271,7 @@ int PackVmlinuzI386::decompressKernel()
             if (fd < 0)
                 break;
             gzFile zf = gzdopen(fd, "rb");
-            if (zf == NULL)
+            if (zf == nullptr)
                 break;
             // estimate gzip-decompressed kernel size & alloc buffer
             if (ibuf.getSize() == 0)
@@ -344,7 +345,7 @@ int PackVmlinuzI386::decompressKernel()
         // some checks
         if (fd_pos != file_size)
         {
-            //printf("fd_pos: %ld, file_size: %ld\n", (long)fd_pos, (long)file_size);
+            NO_printf("fd_pos: %lld, file_size: %lld\n", fd_pos, file_size);
 
             // linux-2.6.21.5/arch/i386/boot/compressed/vmlinux.lds
             // puts .data.compressed ahead of .text, .rodata, etc;
@@ -434,15 +435,15 @@ void PackVmlinuzI386::buildLoader(const Filter *ft)
               ft->id ? "LZCALLT1" : "",
               "LZIMAGE0",
               getDecompressorSections(),
-              NULL
+              nullptr
              );
     if (ft->id)
     {
         assert(ft->calls > 0);
-        addLoader("LZCALLT9", NULL);
+        addLoader("LZCALLT9", nullptr);
         addFilter32(ft->id);
     }
-    addLoader("LINUZ990,IDENTSTR,UPX1HEAD", NULL);
+    addLoader("LINUZ990,IDENTSTR,UPX1HEAD", nullptr);
 }
 
 
@@ -511,7 +512,7 @@ void PackBvmlinuzI386::buildLoader(const Filter *ft)
             "LINUZ140,LZCUTPOI,LINUZ141",
             (ft->id ? "LINUZ145" : ""),
             (ph.first_offset_found == 1 ? "LINUZ010" : ""),
-            NULL);
+            nullptr);
     }
     else {
         addLoader("LINUZ000,LINUZ001,LINUZVGA,LINUZ005",
@@ -521,26 +522,26 @@ void PackBvmlinuzI386::buildLoader(const Filter *ft)
               "+40", // align the stuff to 4 byte boundary
               "UPX1HEAD", // 32 byte
               "LZCUTPOI",
-              NULL);
+              nullptr);
         // fake alignment for the start of the decompressor
         //linker->defineSymbol("LZCUTPOI", 0x1000);
     }
 
-    addLoader(getDecompressorSections(), NULL);
+    addLoader(getDecompressorSections(), nullptr);
 
     if (ft->id)
     {
             assert(ft->calls > 0);
         if (0x40==(0xf0 & ft->id)) {
-            addLoader("LZCKLLT9", NULL);
+            addLoader("LZCKLLT9", nullptr);
         }
         else {
-            addLoader("LZCALLT9", NULL);
+            addLoader("LZCALLT9", nullptr);
         }
         addFilter32(ft->id);
     }
     if (0!=page_offset) {
-        addLoader("LINUZ150,IDENTSTR,+40,UPX1HEAD", NULL);
+        addLoader("LINUZ150,IDENTSTR,+40,UPX1HEAD", nullptr);
         unsigned const l_len = getLoaderSize();
         unsigned const c_len = ALIGN_UP(ph.c_len, 4u);
         unsigned const e_len = getLoaderSectionStart("LINUZ141") -
@@ -558,7 +559,7 @@ void PackBvmlinuzI386::buildLoader(const Filter *ft)
         linker->defineSymbol("unc_offset", ph.overlap_overhead + ph.u_len - c_len);
     }
     else {
-        addLoader("LINUZ990", NULL);
+        addLoader("LINUZ990", nullptr);
     }
 }
 
@@ -594,8 +595,8 @@ void PackBvmlinuzI386::pack(OutputFile *fo)
             (res->lit_context_bits << 0) |
             (res->lit_pos_bits << 8) |
             (res->pos_bits << 16);
-        if (linker->bele->isBE()) // big endian - bswap32
-            acc_swab32s(&properties);
+        if (bele->isBE()) // big endian - bswap32
+            properties = bswap32(properties);
         linker->defineSymbol("lzma_properties", properties);
         // -2 for properties
         linker->defineSymbol("lzma_c_len", ph.c_len - 2);
@@ -670,7 +671,7 @@ void PackBvmlinuzI386::pack(OutputFile *fo)
 // unpack
 **************************************************************************/
 
-int PackVmlinuzI386::canUnpack()
+tribool PackVmlinuzI386::canUnpack()
 {
     if (readFileHeader() != getFormat())
         return false;
@@ -687,7 +688,7 @@ void PackVmlinuzI386::unpack(OutputFile *fo)
     // FIXME: but we could write the uncompressed "vmlinux" image
 
     ibuf.alloc(ph.c_len);
-    obuf.allocForUncompression(ph.u_len);
+    obuf.allocForDecompression(ph.u_len);
 
     fi->seek(setup_size + ph.buf_offset + ph.getPackHeaderSize(), SEEK_SET);
     fi->readx(ibuf, ph.c_len);
@@ -734,7 +735,7 @@ int PackVmlinuzARMEL::getStrategy(Filter &/*ft*/)
     return (opt->no_filter ? -3 : ((opt->filter > 0) ? -2 : 2));
 }
 
-bool PackVmlinuzARMEL::canPack()
+tribool PackVmlinuzARMEL::canPack()
 {
     return readFileHeader() == getFormat();
 }
@@ -749,7 +750,7 @@ int PackVmlinuzARMEL::readFileHeader()
             return 0;
         }
     }
-    return UPX_F_VMLINUZ_ARMEL;
+    return UPX_F_VMLINUZ_ARM;
 }
 
 int PackVmlinuzARMEL::decompressKernel()
@@ -759,7 +760,7 @@ int PackVmlinuzARMEL::decompressKernel()
     fi->seek(0, SEEK_SET);
     fi->readx(obuf, file_size);
 
-    //checkAlreadyPacked(obuf + setup_size, UPX_MIN(file_size - setup_size, (off_t)1024));
+    //checkAlreadyPacked(obuf + setup_size, UPX_MIN(file_size - setup_size, 1024LL));
 
     // Find head.S:
     //      bl decompress_kernel  # 0xeb......
@@ -841,7 +842,7 @@ int PackVmlinuzARMEL::decompressKernel()
         // try to decompress
         int klen;
         int fd;
-        off_t fd_pos;
+        upx_off_t fd_pos;
         for (;;)
         {
             klen = -1;
@@ -852,7 +853,7 @@ int PackVmlinuzARMEL::decompressKernel()
             if (fd < 0)
                 break;
             gzFile zf = gzdopen(fd, "rb");
-            if (zf == NULL)
+            if (zf == nullptr)
                 break;
             // estimate gzip-decompressed kernel size & alloc buffer
             if (ibuf.getSize() == 0)
@@ -920,32 +921,32 @@ Linker* PackVmlinuzARMEL::newLinker() const
     return new ElfLinkerArmLE;
 }
 
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/arm.v5a-linux.kernel.vmlinux.h"
-static const
+static const CLANG_FORMAT_DUMMY_STATEMENT
 #include "stub/arm.v5a-linux.kernel.vmlinuz-head.h"
 
 void PackVmlinuzARMEL::buildLoader(const Filter *ft)
 {
     // prepare loader; same as vmlinux (with 'x')
     initLoader(stub_arm_v5a_linux_kernel_vmlinux, sizeof(stub_arm_v5a_linux_kernel_vmlinux));
-    addLoader("LINUX000", NULL);
+    addLoader("LINUX000", nullptr);
     if (ft->id) {
         assert(ft->calls > 0);
-        addLoader("LINUX010", NULL);
+        addLoader("LINUX010", nullptr);
     }
-    addLoader("LINUX020", NULL);
+    addLoader("LINUX020", nullptr);
     if (ft->id) {
         addFilter32(ft->id);
     }
-    addLoader("LINUX030", NULL);
-         if (ph.method == M_NRV2E_8) addLoader("NRV2E", NULL);
-    else if (ph.method == M_NRV2B_8) addLoader("NRV2B", NULL);
-    else if (ph.method == M_NRV2D_8) addLoader("NRV2D", NULL);
+    addLoader("LINUX030", nullptr);
+         if (ph.method == M_NRV2E_8) addLoader("NRV2E", nullptr);
+    else if (ph.method == M_NRV2B_8) addLoader("NRV2B", nullptr);
+    else if (ph.method == M_NRV2D_8) addLoader("NRV2D", nullptr);
     else if (M_IS_LZMA(ph.method))   addLoader("LZMA_ELF00",
-        (opt->small ? "LZMA_DEC10" : "LZMA_DEC20"), "LZMA_DEC30", NULL);
+        (opt->small ? "LZMA_DEC10" : "LZMA_DEC20"), "LZMA_DEC30", nullptr);
     else throwBadLoader();
-    addLoader("IDENTSTR,UPX1HEAD", NULL);
+    addLoader("IDENTSTR,UPX1HEAD", nullptr);
 
     // To debug (2008-09-14):
     //   Build gdb-6.8-21.fc9.src.rpm; ./configure --target=arm-none-elf; make
@@ -1028,7 +1029,7 @@ void PackVmlinuzARMEL::pack(OutputFile *fo)
         throwNotCompressible();
 }
 
-int PackVmlinuzARMEL::canUnpack()
+tribool PackVmlinuzARMEL::canUnpack()
 {
     if (readFileHeader() != getFormat())
         return false;
@@ -1044,7 +1045,7 @@ void PackVmlinuzARMEL::unpack(OutputFile *fo)
     // FIXME: but we could write the uncompressed "vmlinux" image
 
     ibuf.alloc(ph.c_len);
-    obuf.allocForUncompression(ph.u_len);
+    obuf.allocForDecompression(ph.u_len);
 
     fi->seek(setup_size + ph.buf_offset + ph.getPackHeaderSize(), SEEK_SET);
     fi->readx(ibuf, ph.c_len);
