@@ -2,9 +2,9 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2018 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2018 Laszlo Molnar
-   Copyright (C) 2000-2018 John F. Reiser
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
+   Copyright (C) 2000-2024 John F. Reiser
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -128,10 +128,19 @@ xread(Extent *x, void *buf, size_t count)
     for (j = count; 0!=j--; ++p, ++q) {
         *q = *p;
     }
+    DPRINTF("   buf: %%x %%x %%x\\n", ((int *)buf)[0], ((int *)buf)[1], ((int *)buf)[2]);
     x->buf  += count;
     x->size -= count;
 }
 
+static void
+xpeek(Extent *x, void *buf, size_t count)
+{
+    DPRINTF("xpeek buf=%%p  count=%%x  ", buf, count);
+    xread(x, buf, count);
+    x->size += count;
+    x->buf  -= count;
+}
 
 /*************************************************************************
 // util
@@ -176,7 +185,7 @@ struct b_info { // 12-byte header before each compressed block
     unsigned char b_method;  // compression algorithm
     unsigned char b_ftid;  // filter id
     unsigned char b_cto8;  // filter parameter
-    unsigned char b_unused;
+    unsigned char b_extra;
 };
 
 typedef void f_unfilter(
@@ -310,7 +319,8 @@ typedef struct {
             MH_MAGIC64 = 1+0xfeedface
         };
         enum e2 {
-            MH_EXECUTE = 2
+            MH_EXECUTE = 2,
+            MH_DYLINKER= 7     /* /usr/bin/dyld */
         };
         enum e3 {
             MH_NOUNDEFS = 1
@@ -453,6 +463,7 @@ typedef union {
 #endif  //}
 #define MAP_ANON_FD    -1
 
+#define PROT_NONE      0
 #define PROT_READ      1
 #define PROT_WRITE     2
 #define PROT_EXEC      4
@@ -461,7 +472,78 @@ extern void *mmap(void *, size_t, unsigned, unsigned, int, off_t_upx_stub);
 ssize_t pread(int, void *, size_t, off_t_upx_stub);
 extern void bswap(void *, unsigned);
 
-// FIXME: must reserve convex hull of pages; must reloc PIE
+typedef size_t Addr;  // this source file is used by 32-bit and 64-bit machines
+
+void my_bkpt(void const *foo, ...)
+{
+    (void)foo;
+#if defined(__x86_64__)  //{
+    __asm__(".byte 0xcc");
+#elif  defined(__aarch64__)  //}{
+     __asm__("brk #0");
+#elif  defined(__arm__)  //}{
+     __asm__(".long 0xe7f001f0"); // "bkpt" ==> "udf #16"
+#endif  //}
+}
+
+// Find convex hull of PT_LOAD (the minimal interval which covers all PT_LOAD),
+// and mmap that much, to be sure that a kernel using exec-shield-randomize
+// won't place the first piece in a way that leaves no room for the rest.
+static Addr // returns relocation constant
+xfind_pages(
+    Mach_header const *const mhdr,
+    Mach_segment_command const *sc,
+    int const ncmds,
+    Addr addr
+)
+{
+    Addr lo= ~(Addr)0, hi= 0;
+    int j;
+    unsigned mflags = ((mhdr->filetype == MH_DYLINKER || mhdr->flags & MH_PIE) ? 0 : MAP_FIXED);
+    mflags += MAP_PRIVATE | MAP_ANON;  // '+' can optimize better than '|'
+    DPRINTF("xfind_pages  mhdr=%%p  sc=%%p  ncmds=%%d  addr=%%p  mflags=%%x\\n",
+        mhdr, sc, ncmds, addr, mflags);
+    for (j=0; j < ncmds; ++j,
+        (sc = (Mach_segment_command const *)((sc->cmdsize>>2) + (unsigned const *)sc))
+    ) if (LC_SEGMENT==sc->cmd) {
+        DPRINTF("  #%%d  cmd=%%x  cmdsize=%%x  vmaddr=%%p  vmsize==%%p  filesize=%%p  lo=%%p  mflags=%%x\\n",
+            j, sc->cmd, sc->cmdsize, sc->vmaddr, sc->vmsize, sc->filesize, lo, mflags);
+        if (sc->vmsize  // theoretically occupies address space
+        &&  !(sc->vmaddr==0 && (MAP_FIXED & mflags))  // but ignore PAGEZERO when MAP_FIXED
+        ) {
+            if (mhdr->filetype == MH_DYLINKER  // /usr/lib/dyld
+            &&  0==(1+ lo)  // 1st LC_SEGMENT
+            &&  sc->vmaddr != 0  // non-floating address
+            ) {
+                // "pre-linked" dyld on MacOS 10.11.x El Capitan
+                mflags |= MAP_FIXED;
+            }
+            if (lo > sc->vmaddr) {
+                lo = sc->vmaddr;
+            }
+            if (hi < (sc->vmsize + sc->vmaddr)) {
+                hi =  sc->vmsize + sc->vmaddr;
+            }
+        }
+    }
+    lo -= ~PAGE_MASK & lo;  // round down to page boundary
+    hi  =  PAGE_MASK & (hi - lo - PAGE_MASK -1);  // page length
+    DPRINTF("  addr=%%p  lo=%%p  len=%%p  mflags=%%x\\n", addr, lo, hi, mflags);
+    if (MAP_FIXED & mflags) {
+        addr = lo;
+        int rv = munmap((void *)addr, hi);
+        if (rv) {
+            DPRINTF("munmap addr=%%p len=%%p, rv=%%x\\n", addr, hi, rv);
+        }
+    }
+    addr = (Addr)mmap((void *)addr, hi, PROT_NONE, mflags, MAP_ANON_FD, 0);
+    DPRINTF("  addr=%%p\\n", addr);
+    if (~PAGE_MASK & addr) {
+        err_exit(6);
+    }
+    return (Addr)(addr - lo);
+}
+
 unsigned * // &hatch if main; &Mach_thread_state if dyld
 do_xmap(
     Mach_header *const mhdr,
@@ -473,112 +555,147 @@ do_xmap(
     f_unfilter *const f_unf
 )
 {
-    unsigned *rv = 0;
-    Mach_segment_command *sc = (Mach_segment_command *)(1+ mhdr);
-    size_t reloc = 0;
-    unsigned j;
-
     DPRINTF("do_xmap  fdi=%%x  mhdr=%%p  *mhdrpp=%%p  xi=%%p(%%x %%p) f_unf=%%p\\n",
         fdi, mhdr, (mhdrpp ? *mhdrpp : 0), xi, (xi? xi->size: 0), (xi? xi->buf: 0), f_unf);
 
+    unsigned *rv = 0;
+    Extent xi_orig = {0, 0};
+    Mach_segment_command *sc = (Mach_segment_command *)(1+ mhdr);
+    Addr const reloc = xfind_pages(mhdr, sc, mhdr->ncmds, 0);
+    DPRINTF("do_xmap reloc=%%p\\n", reloc);
+    unsigned j;
+    if (xi) { // remember "Beginning of tape"
+        xi_orig = *xi;
+    }
     for ( j=0; j < mhdr->ncmds; ++j,
         (sc = (Mach_segment_command *)((sc->cmdsize>>2) + (unsigned *)sc))
     ) {
-        DPRINTF("  #%%d  cmd=%%x  cmdsize=%%x\\n", j, sc->cmd, sc->cmdsize);
-        if (LC_SEGMENT==sc->cmd && sc->vmsize) {
-            if (!sc->fileoff && sc->filesize  // should be __TEXT
-            && *mhdrpp && ((*mhdrpp)->flags & MH_PIE)) {
-                reloc = (size_t)*mhdrpp - sc->vmaddr;
-                DPRINTF("reloc=%%p\\n", reloc);
+        DPRINTF("  #%%d  sc=%%x  cmd=%%x  cmdsize=%%x  vmsize=%%x\\n",
+                j, sc, sc->cmd, sc->cmdsize, sc->vmsize);
+        if (LC_SEGMENT==sc->cmd) {
+            struct b_info h;
+            if (xi && sc->filesize) { // Find the correct compressed block.
+                xpeek(xi, (unsigned char *)&h, sizeof(h));
+                DPRINTF("  h.b_extra=%%d  j=%%d\n", h.b_extra, j);
+                if (h.b_extra != j) { // not the next one
+                    *xi = xi_orig;  // rewind
+                    for (;;) {
+                        xpeek(xi, (unsigned char *)&h, sizeof(h));
+                        if (h.b_extra == j) {
+                            break;
+                        }
+                        xi->size -= sizeof(h) + h.sz_cpr;
+                        xi->buf  += sizeof(h) + h.sz_cpr;
+                    }
+                }
             }
-            Extent xo;
-            size_t mlen = xo.size = sc->filesize;
-            unsigned char  *addr = xo.buf = reloc + (unsigned char *)sc->vmaddr;
-            unsigned char *haddr = sc->vmsize + addr;
-            size_t frag = (int)(uint64_t)addr &~ PAGE_MASK;
-            addr -= frag;
-            mlen += frag;
+            if (!sc->vmsize) { // not mapped, such as __DWARF info for 'go'
+                if (xi) {
+                    DPRINTF("    0==.vmsize; skipping %%x\\n", h.sz_cpr);
+                    xi->size -= sizeof(h) + h.sz_cpr;
+                    xi->buf  += sizeof(h) + h.sz_cpr;
+                }
+                continue;  // redundant for clarity
+            }
+            else { // finally, some meat!
+                Extent xo;
+                size_t mlen = xo.size = sc->filesize;
+                              xo.buf  = (void *)(reloc + sc->vmaddr);
+                Addr  addr = (Addr)xo.buf;
+                Addr haddr = sc->vmsize + addr;
+                size_t frag = addr &~ PAGE_MASK;
+                addr -= frag;
+                mlen += frag;
 
-            if (0!=mlen) {
-                // Decompressor can overrun the destination by 3 bytes.  [x86 only]
-                size_t const mlen3 = mlen + (xi ? 3 : 0);
-                unsigned const prot = VM_PROT_READ | VM_PROT_WRITE;
-                unsigned const flags = (*mhdrpp ? MAP_FIXED : 0) | MAP_PRIVATE |
-                            ((xi || 0==sc->filesize) ? MAP_ANON : 0);
-                int const fdm = ((0==sc->filesize) ? MAP_ANON_FD : fdi);
-                off_t_upx_stub const offset = sc->fileoff + fat_offset;
+                DPRINTF("    mlen=%%p  frag=%%p  addr=%%p\\n", mlen, frag, addr);
+                if (0!=mlen) {
+                    size_t const mlen3 = mlen
+        #if defined(__x86_64__)  //{
+                        // Decompressor can overrun the destination by 3 bytes.  [x86 only]
+                        + (xi ? 3 : 0)
+        #endif  //}
+                        ;
+                    unsigned const prot = VM_PROT_READ | VM_PROT_WRITE;
+                    // MAP_FIXED: xfind_pages() reserved them, so use them!
+                    unsigned const flags = MAP_FIXED | MAP_PRIVATE |
+                                    ((xi || 0==sc->filesize) ? MAP_ANON    : 0);
+                    int const fdm = ((xi || 0==sc->filesize) ? MAP_ANON_FD : fdi);
+                    off_t_upx_stub const offset = sc->fileoff + fat_offset;
 
-                DPRINTF("mmap  addr=%%p  len=%%p  prot=%%x  flags=%%x  fd=%%d  off=%%p  reloc=%%p\\n",
-                    addr, mlen3, prot, flags, fdm, offset, reloc);
-                unsigned *maddr = mmap(addr, mlen3, prot, flags, fdm, offset);
-                if (!*mhdrpp) { // MH_DYLINKER
-                    *mhdrpp = (Mach_header*)maddr;
-                    addr = (unsigned char *)maddr;
-                    reloc = (size_t)addr;
+                    DPRINTF("mmap  addr=%%p  len=%%p  prot=%%x  flags=%%x  fd=%%d  off=%%p  reloc=%%p\\n",
+                        addr, mlen3, prot, flags, fdm, offset, reloc);
+                    {
+                        Addr maddr = (Addr)mmap((void *)addr, mlen3, prot, flags, fdm, offset);
+                        DPRINTF("maddr=%%p\\n", maddr);
+                        if (maddr != addr) {
+                            err_exit(8);
+                        }
+                        addr = maddr;
+                    }
+                    if (!*mhdrpp) { // MH_DYLINKER
+                        *mhdrpp = (Mach_header*)addr;
+                    }
                 }
-                if (addr != (unsigned char *)maddr) {
-                    DPRINTF("maddr=%%p  addr=%%p\\n", maddr, addr);
-                    err_exit(8);
+                if (xi && 0!=sc->filesize) {
+                    if (0==sc->fileoff /*&& 0!=mhdrpp*/) {
+                        *mhdrpp = (Mach_header *)(void *)addr;
+                    }
+                    unpackExtent(xi, &xo, f_exp, f_unf);
                 }
-            }
-            if (xi && 0!=sc->filesize) {
-                if (0==sc->fileoff /*&& 0!=mhdrpp*/) {
-                    *mhdrpp = (Mach_header *)(void *)addr;
+                DPRINTF("xi=%%p  mlen=%%p  fileoff=%%p  nsects=%%d\\n",
+                    xi, mlen, sc->fileoff, sc->nsects);
+                if (xi && mlen && !sc->fileoff && sc->nsects) {
+                    // main target __TEXT segment at beginning of file with sections (__text)
+                    // Use upto 2 words of header padding for the escape hatch.
+                    // fold.S could do this easier, except PROT_WRITE is missing then.
+                    union {
+                        unsigned char  *p0;
+                        unsigned short *p1;
+                        unsigned int   *p2;
+                        unsigned long  *p3;
+                    } u;
+                    u.p0 = (unsigned char *)addr;
+                    Mach_segment_command *segp = (Mach_segment_command *)((((char *)sc - (char *)mhdr)>>2) + u.p2);
+                    Mach_section_command *const secp = (Mach_section_command *)(1+ segp);
+        #if defined(__aarch64__)  //{
+                    unsigned *hatch= -2+ (secp->offset>>2) + u.p2;
+                    hatch[0] = 0xd4000001;  // svc #0  // syscall
+                    hatch[1] = 0xd65f03c0;  // ret
+        #elif defined(__arm__)  //}{
+                    unsigned *hatch= -2+ (secp->offset>>2) + u.p2;
+                    hatch[0] = 0xef000000;  // svc 0x0  // syscall
+                    hatch[1] = 0xe12fff1e;  // bx lr
+        #elif defined(__x86_64__)  //}{
+                    unsigned *hatch= -1+ (secp->offset>>2) + u.p2;
+                    hatch[0] = 0xc3050f90;  // nop; syscall; ret
+        #endif  //}
+                    DPRINTF("hatch=%%p  secp=%%p  segp=%%p  mhdr=%%p\\n", hatch, secp, segp, addr);
+                    rv = hatch;
                 }
-                unpackExtent(xi, &xo, f_exp, f_unf);
-            }
-            DPRINTF("xi=%%p  mlen=%%p  fileoff=%%p  nsects=%%d\\n",
-                xi, mlen, sc->fileoff, sc->nsects);
-            if (xi && mlen && !sc->fileoff && sc->nsects) {
-                // main target __TEXT segment at beginning of file with sections (__text)
-                // Use upto 2 words of header padding for the escape hatch.
-                // fold.S could do this easier, except PROT_WRITE is missing then.
-                union {
-                    unsigned char  *p0;
-                    unsigned short *p1;
-                    unsigned int   *p2;
-                    unsigned long  *p3;
-                } u;
-                u.p0 = addr;
-                Mach_segment_command *segp = (Mach_segment_command *)((((char *)sc - (char *)mhdr)>>2) + u.p2);
-                Mach_section_command *const secp = (Mach_section_command *)(1+ segp);
-    #if defined(__aarch64__)  //{
-                unsigned *hatch= -2+ (secp->offset>>2) + u.p2;
-                hatch[0] = 0xd4000001;  // svc #0  // syscall
-                hatch[1] = 0xd65f03c0;  // ret
-    #elif defined(__arm__)  //}{
-                unsigned *hatch= -2+ (secp->offset>>2) + u.p2;
-                hatch[0] = 0xef000000;  // svc 0x0  // syscall
-                hatch[1] = 0xe12fff1e;  // bx lr
-    #elif defined(__x86_64__)  //}{
-                unsigned *hatch= -1+ (secp->offset>>2) + u.p2;
-                hatch[0] = 0xc3050f90;  // nop; syscall; ret
-    #endif  //}
-                DPRINTF("hatch=%%p  secp=%%p  segp=%%p  mhdr=%%p\\n", hatch, secp, segp, addr);
-                rv = hatch;
-            }
-            /*bzero(addr, frag);*/  // fragment at lo end
-            frag = (-mlen) &~ PAGE_MASK;  // distance to next page boundary
-            bzero(mlen+addr, frag);  // fragment at hi end
-            if (0!=mlen && 0!=mprotect(addr, mlen, sc->initprot)) {
-                err_exit(10);
-    ERR_LAB
-            }
-            addr += mlen + frag;  /* page boundary on hi end */
-            if (
-    #if SIMULATE_ON_LINUX_EABI4  /*{*/
-                0!=addr &&
-    #endif  /*}*/
-                            addr < haddr) { // need pages for .bss
-                if (0!=addr && addr != mmap(addr, haddr - addr, sc->initprot,
-                        MAP_FIXED | MAP_PRIVATE | MAP_ANON, MAP_ANON_FD, 0 ) ) {
-                    err_exit(9);
+                /*bzero(addr, frag);*/  // fragment at lo end
+                frag = (-mlen) &~ PAGE_MASK;  // distance to next page boundary
+                bzero((void *)(mlen+addr), frag);  // fragment at hi end
+                if (0!=mlen && 0!=mprotect((void *)addr, mlen, sc->initprot)) {
+                    err_exit(10);
+        ERR_LAB
                 }
-            }
-            else if (xi) { // cleanup if decompressor overrun crosses page boundary
-                mlen = ~PAGE_MASK & (3+ mlen);
-                if (mlen<=3) { // page fragment was overrun buffer only
-                    munmap((char *)addr, mlen);
+                addr += mlen + frag;  /* page boundary on hi end */
+                if (
+        #if SIMULATE_ON_LINUX_EABI4  /*{*/
+                    0!=addr &&
+        #endif  /*}*/
+                                addr < haddr) { // need pages for .bss
+                    if (0!=addr && addr != (Addr)mmap((void *)addr, haddr - addr, sc->initprot,
+                            MAP_FIXED | MAP_PRIVATE | MAP_ANON, MAP_ANON_FD, 0 ) ) {
+                        err_exit(9);
+                    }
+                }
+                else if (xi) { // cleanup if decompressor overrun crosses page boundary
+                    mlen = ~PAGE_MASK & (3+ mlen);
+                    if (mlen<=3) { // page fragment was overrun buffer only
+                        DPRINTF("munmap  %%x  %%x\\n", addr, mlen);
+                        munmap((char *)addr, mlen);
+                    }
                 }
             }
         }
@@ -607,7 +724,6 @@ do_xmap(
     return rv;
 }
 
-
 /*************************************************************************
 // upx_main - called by our unfolded entry code
 //
@@ -659,6 +775,7 @@ upx_main(
             err_exit(18);
         }
 fat:
+        DPRINTF("fat_offset= %%p\\n", fat_offset);
         if ((ssize_t)sz_mhdr!=pread(fdi, (void *)mhdr, sz_mhdr, fat_offset)) {
 ERR_LAB
             err_exit(19);
@@ -808,7 +925,7 @@ finish:
 done:
     va_end(va);
     return n;
- }
+}
 
 #endif  //}
 
