@@ -2,9 +2,9 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2018 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2018 Laszlo Molnar
-   Copyright (C) 2000-2018 John F. Reiser
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
+   Copyright (C) 2000-2024 John F. Reiser
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -46,65 +46,25 @@
 // static data, and no string constants.
 
 #if !DEBUG  /*{*/
-#define DPRINTF(a) /* empty: no debug drivel */
-#define DEBUG_STRCON(name, value) /* empty */
+#define DPRINTF(a ...) /* empty: no debug drivel */
 #else  /*}{ DEBUG */
-extern int write(int, void const *, size_t);
-#if 0
-#include "stdarg.h"
-#else
+#define DPRINTF(fmt, args...) ({ \
+    char const *r_fmt; \
+    asm("call 0f; .asciz \"" fmt "\"; 0: pop %0" \
+/*out*/ : "=r"(r_fmt) ); \
+    dprintf(r_fmt, args); \
+})
+
 #define va_arg      __builtin_va_arg
 #define va_end      __builtin_va_end
 #define va_list     __builtin_va_list
 #define va_start    __builtin_va_start
-#endif
-
-#if defined(__i386__) || defined(__x86_64__) /*{*/
-#define PIC_STRING(value, var) \
-    __asm__ __volatile__ ( \
-        "call 0f; .asciz \"" value "\"; \
-      0: pop %0;" : "=r"(var) : \
-    )
-#elif defined(__arm__)  /*}{*/
-#define PIC_STRING(value, var) \
-    __asm__ __volatile__ ( \
-        "mov %0,pc; b 0f; \
-        .asciz \"" value "\"; .balign 4; \
-      0: " : "=r"(var) \
-    )
-#elif defined(__mips__)  /*}{*/
-#define PIC_STRING(value, var) \
-    __asm__ __volatile__ ( \
-        ".set noreorder; bal 0f; move %0,$31; .set reorder; \
-        .asciz \"" value "\"; .balign 4; \
-      0: " \
-        : "=r"(var) : : "ra" \
-    )
-#endif  /*}*/
-
-
-#define DEBUG_STRCON(name, strcon) \
-    static char const *name(void) { \
-        register char const *rv; PIC_STRING(strcon, rv); \
-        return rv; \
-    }
-
-
-#ifdef __arm__  /*{*/
-extern unsigned div10(unsigned);
-#else  /*}{*/
-static unsigned
-div10(unsigned x)
-{
-    return x / 10u;
-}
-#endif  /*}*/
 
 static int
 unsimal(unsigned x, char *ptr, int n)
 {
     if (10<=x) {
-        unsigned const q = div10(x);
+        unsigned const q = x / 10;
         x -= 10 * q;
         n = unsimal(q, ptr, n);
     }
@@ -122,54 +82,80 @@ decimal(int x, char *ptr, int n)
     return unsimal(x, ptr, n);
 }
 
-DEBUG_STRCON(STR_hex, "0123456789abcdef");
-
 static int
 heximal(unsigned long x, char *ptr, int n)
 {
-    if (16<=x) {
-        n = heximal(x>>4, ptr, n);
-        x &= 0xf;
+    unsigned j = -1+ 2*sizeof(unsigned long);
+    unsigned long m = 0xful << (4 * j);
+    for (; j; --j, m >>= 4) { // omit leading 0 digits
+        if (m & x) break;
     }
-    ptr[n] = STR_hex()[x];
-    return 1+ n;
+    for (; m; --j, m >>= 4) {
+        unsigned d = 0xf & (x >> (4 * j));
+        ptr[n++] = ((10<=d) ? ('a' - 10) : '0') + d;
+    }
+    return n;
 }
 
-
-#define DPRINTF(a) dprintf a
-
 static int
 dprintf(char const *fmt, ...)
 {
-    char c;
     int n= 0;
-    char *ptr;
-    char buf[20];
+    char const *literal = 0;  // NULL
+    char buf[24];  // ~0ull == 18446744073709551615 ==> 20 chars
     va_list va; va_start(va, fmt);
-    ptr= &buf[0];
-    while (0!=(c= *fmt++)) if ('%'!=c) goto literal;
-    else switch (c= *fmt++) {
-    default: {
-literal:
-        n+= write(2, fmt-1, 1);
-    } break;
-    case 0: goto done;  /* early */
-    case 'u': {
-        n+= write(2, buf, unsimal(va_arg(va, unsigned), buf, 0));
-    } break;
-    case 'd': {
-        n+= write(2, buf, decimal(va_arg(va, int), buf, 0));
-    } break;
-    case 'p': {
-        buf[0] = '0';
-        buf[1] = 'x';
-        n+= write(2, buf, heximal((unsigned long)va_arg(va, void *), buf, 2));
-    } break;
-    case 'x': {
-        buf[0] = '0';
-        buf[1] = 'x';
-        n+= write(2, buf, heximal(va_arg(va, int), buf, 2));
-    } break;
+    for (;;) {
+        char c = *fmt++;
+        if (!c) { // end of fmt
+            if (literal) {
+                goto finish;
+            }
+            break;  // goto done
+        }
+        if ('%'!=c) {
+            if (!literal) {
+                literal = fmt;  // 1 beyond start of literal
+            }
+            continue;
+        }
+        // '%' == c
+        if (literal) {
+finish:
+            n += write(2, -1+ literal, fmt - literal);
+            literal = 0;  // NULL
+            if (!c) { // fmt already ended
+               break;  // goto done
+            }
+        }
+        switch (c= *fmt++) { // deficiency: does not handle _long_
+        default: { // un-implemented conversion
+            n+= write(2, -1+ fmt, 1);
+        } break;
+        case 0: { // fmt ends with "%\0" ==> ignore
+            goto done;
+        } break;
+        case 'u': {
+            n+= write(2, buf, unsimal(va_arg(va, unsigned), buf, 0));
+        } break;
+        case 'd': {
+            n+= write(2, buf, decimal(va_arg(va, int), buf, 0));
+        } break;
+        case 'p': {
+            buf[0] = '0';
+            buf[1] = 'x';
+            n+= write(2, buf, heximal((unsigned long)va_arg(va, void *), buf, 2));
+        } break;
+        case 'x': {
+            buf[0] = '0';
+            buf[1] = 'x';
+            n+= write(2, buf, heximal(va_arg(va, unsigned int), buf, 2));
+        } break;
+        case 's': {
+            char *s0= (char *)va_arg(va, unsigned char *), *s= s0;
+            if (s) while (*s) ++s;
+            n+= write(2, s0, s - s0);
+        } break;
+        } // 'switch'
     }
 done:
     va_end(va);
@@ -177,6 +163,7 @@ done:
 }
 #endif  /*}*/
 
+extern int spin(int);
 
 /*************************************************************************
 // "file" util
@@ -187,17 +174,16 @@ typedef struct {
     void *buf;
 } Extent;
 
-DEBUG_STRCON(STR_xread, "xread %%p(%%x %%p) %%p %%x\\n")
-DEBUG_STRCON(STR_xreadfail, "xreadfail %%p(%%x %%p) %%p %%x\\n")
-
 static void
 xread(Extent *x, void *buf, size_t count)
 {
     unsigned char *p=x->buf, *q=buf;
     size_t j;
-    DPRINTF((STR_xread(), x, x->size, x->buf, buf, count));
+    DPRINTF("xread %%p(%%x %%p) %%p %%x\\n",
+            x, x->size, x->buf, buf, count);
     if (x->size < count) {
-        DPRINTF((STR_xreadfail(), x, x->size, x->buf, buf, count));
+        DPRINTF("xreadfail %%p(%%x %%p) %%p %%x\\n",
+                x, x->size, x->buf, buf, count);
         exit(127);
     }
     for (j = count; 0!=j--; ++p, ++q) {
@@ -217,7 +203,6 @@ xread(Extent *x, void *buf, size_t count)
 #define err_exit(a) goto error
 #else  //}{  save debugging time
 #define ERR_LAB /*empty*/
-DEBUG_STRCON(STR_exit, "err_exit %%x\\n");
 
 static void
 err_exit(int a)
@@ -265,10 +250,6 @@ typedef int f_expand(
     const nrv_byte *, nrv_uint,
           nrv_byte *, nrv_uint *, unsigned );
 
-DEBUG_STRCON(STR_unpackExtent,
-        "unpackExtent in=%%p(%%x %%p)  out=%%p(%%x %%p)  %%p %%p\\n");
-DEBUG_STRCON(STR_err5, "sz_cpr=%%x  sz_unc=%%x  xo->size=%%x\\n");
-
 static void
 unpackExtent(
     Extent *const xi,  // input
@@ -277,8 +258,8 @@ unpackExtent(
     f_unfilter *f_unf
 )
 {
-    DPRINTF((STR_unpackExtent(),
-        xi, xi->size, xi->buf, xo, xo->size, xo->buf, f_decompress, f_unf));
+    DPRINTF("unpackExtent in=%%p(%%x %%p)  out=%%p(%%x %%p)  %%p %%p\\n",
+        xi, xi->size, xi->buf, xo, xo->size, xo->buf, f_decompress, f_unf);
     while (xo->size) {
         struct b_info h;
         //   Note: if h.sz_unc == h.sz_cpr then the block was not
@@ -286,6 +267,8 @@ unpackExtent(
 
         // Read and check block sizes.
         xread(xi, (unsigned char *)&h, sizeof(h));
+        DPRINTF("  sz_unc=%%x  sz_cpr=%%x  param=%%x\\n",
+            h.sz_unc, h.sz_cpr, *(int *)&h.b_method);
         if (h.sz_unc == 0) {                     // uncompressed size 0 -> EOF
             if (h.sz_cpr != UPX_MAGIC_LE32)      // h.sz_cpr must be h->magic
                 err_exit(2);
@@ -299,7 +282,8 @@ ERR_LAB
         }
         if (h.sz_cpr > h.sz_unc
         ||  h.sz_unc > xo->size ) {
-            DPRINTF((STR_err5(), h.sz_cpr, h.sz_unc, xo->size));
+            DPRINTF("sz_cpr=%%x  sz_unc=%%x  xo->size=%%x\\n",
+                        h.sz_cpr, h.sz_unc, xo->size);
             err_exit(5);
         }
         // Now we have:
@@ -313,7 +297,10 @@ ERR_LAB
                 xo->buf, &out_len, h.b_method);
             if (j != 0 || out_len != (nrv_uint)h.sz_unc)
                 err_exit(7);
+            DPRINTF("  b_ftid=%%x  f_unf=%%p\\n", h.b_ftid, f_unf);
             if (h.b_ftid!=0 && f_unf) {  // have filter
+                DPRINTF(" unfiltering f_unf=%%p  buf=%%p  len=%%x  cto=%%x  ftid=%%x\\n",
+                        f_unf, xo->buf, out_len, h.b_cto8, h.b_ftid);
                 (*f_unf)(xo->buf, out_len, h.b_cto8, h.b_ftid);
             }
             xi->buf  += h.sz_cpr;
@@ -349,6 +336,8 @@ upx_bzero(unsigned char *p, size_t len)
          |(REP8(PROT_WRITE) & EXP8(PF_W)) \
         ) >> ((pf & (PF_R|PF_W|PF_X))<<2) ))
 
+typedef size_t Addr;
+
 typedef struct {
     unsigned magic;
     unsigned nfat_arch;
@@ -408,8 +397,8 @@ typedef struct {
     unsigned cmd;
     unsigned cmdsize;
     char segname[16];
-    unsigned vmaddr;
-    unsigned vmsize;
+    Addr vmaddr;
+    Addr vmsize;
     unsigned fileoff;
     unsigned filesize;
     unsigned maxprot;
@@ -418,11 +407,26 @@ typedef struct {
     unsigned flags;
 } Mach_segment_command;
         enum e5 {
+            VM_PROT_NONE = 0,
             VM_PROT_READ = 1,
             VM_PROT_WRITE = 2,
             VM_PROT_EXECUTE = 4
         };
 
+typedef struct {
+    char sectname[16];
+    char segname[16];
+    Addr addr;   /* memory address */
+    Addr size;   /* size in bytes */
+    unsigned offset; /* file offset */
+    unsigned align;  /* power of 2 */
+    unsigned reloff; /* file offset of relocation entries */
+    unsigned nreloc; /* number of relocation entries */
+    unsigned flags;  /* section type and attributes */
+    unsigned reserved1;  /* for offset or index */
+    unsigned reserved2;  /* for count or sizeof */
+} Mach_section_command;
+
 typedef struct {
     unsigned eax, ebx, ecx, edx;
     unsigned edi, esi, ebp;
@@ -465,102 +469,235 @@ extern void *mmap(void *, size_t, unsigned, unsigned, int, off_t_upx_stub, unsig
 extern ssize_t pread(int, void *, size_t, off_t_upx_stub, unsigned);
 extern void bswap(void *, unsigned);
 
-DEBUG_STRCON(STR_mmap,
-    "mmap  addr=%%p  len=%%p  prot=%%x  flags=%%x  fd=%%d  off=%%p\\n");
-DEBUG_STRCON(STR_do_xmap,
-    "do_xmap  fdi=%%x  mhdr=%%p  xi=%%p(%%x %%p) f_unf=%%p\\n")
+enum {
+        MH_DYLINKER= 7,     /* /usr/bin/dyld */
+        MH_PIE      = 0x200000   // ASLR
 
-static Mach_i386_thread_state const *
-do_xmap(
+};
+
+// Find convex hull of PT_LOAD (the minimal interval which covers all PT_LOAD),
+// and mmap that much, to be sure that a kernel using exec-shield-randomize
+// won't place the first piece in a way that leaves no room for the rest.
+static Addr // returns relocation constant
+xfind_pages(
     Mach_header const *const mhdr,
+    Mach_segment_command const *sc,
+    int const ncmds,
+    Addr addr
+)
+{
+    Addr lo= ~(Addr)0, hi= 0;
+    int j;
+    unsigned mflags = ((mhdr->filetype == MH_DYLINKER || mhdr->flags & MH_PIE) ? 0 : MAP_FIXED);
+    mflags += MAP_PRIVATE | MAP_ANON;  // '+' can optimize better than '|'
+    DPRINTF("xfind_pages  mhdr=%%p  sc=%%p  ncmds=%%d  addr=%%p  mflags=%%x\\n",
+        mhdr, sc, ncmds, addr, mflags);
+    for (j=0; j < ncmds; ++j,
+        (sc = (Mach_segment_command const *)((sc->cmdsize>>2) + (unsigned const *)sc))
+    ) if (LC_SEGMENT==sc->cmd) {
+        DPRINTF("  #%%d  cmd=%%x  cmdsize=%%x  vmaddr=%%p  vmsize==%%p  lo=%%p  mflags=%%x\\n",
+            j, sc->cmd, sc->cmdsize, sc->vmaddr, sc->vmsize, lo, mflags);
+        if (sc->vmsize  // theoretically occupies address space
+        &&  !(sc->vmaddr==0 && (MAP_FIXED & mflags))  // but ignore PAGEZERO when MAP_FIXED
+        ) {
+            if (mhdr->filetype == MH_DYLINKER  // /usr/lib/dyld
+            &&  0==(1+ lo)  // 1st LC_SEGMENT
+            &&  sc->vmaddr != 0  // non-floating address
+            ) {
+                // "pre-linked" dyld on MacOS 10.11.x El Capitan
+                mflags |= MAP_FIXED;
+            }
+            if (lo > sc->vmaddr) {
+                lo = sc->vmaddr;
+            }
+            if (hi < (sc->vmsize + sc->vmaddr)) {
+                hi =  sc->vmsize + sc->vmaddr;
+            }
+        }
+    }
+    lo -= ~PAGE_MASK & lo;  // round down to page boundary
+    hi  =  PAGE_MASK & (hi - lo - PAGE_MASK -1);  // page length
+    DPRINTF("  addr=%%p  lo=%%p  len=%%p  mflags=%%x\\n", addr, lo, hi, mflags);
+    if (MAP_FIXED & mflags) {
+        addr = lo;
+        int rv = munmap((void *)addr, hi);
+        if (rv) {
+            DPRINTF("munmap addr=%%p len=%%p, rv=%%x\\n", addr, hi, rv);
+        }
+    }
+    addr = (Addr)mmap((void *)addr, hi, VM_PROT_NONE, mflags, MAP_ANON_FD, 0, 0);
+    DPRINTF("  addr=%%p\\n", addr);
+    if (~PAGE_MASK & addr) {
+        //err_exit(6);
+    }
+    return (Addr)(addr - lo);
+}
+Addr  // entry: &hatch if main; Mach_thread_state->eip if dyld
+do_xmap(
+    Mach_header *const mhdr,
     off_t_upx_stub const fat_offset,
     Extent *const xi,
     int const fdi,
     Mach_header **mhdrpp,
-    f_expand *const f_decompress,
+    f_expand *const f_exp,
     f_unfilter *const f_unf
 )
 {
-    Mach_segment_command const *sc = (Mach_segment_command const *)(1+ mhdr);
-    Mach_i386_thread_state const *entry = 0;
+    DPRINTF("do_xmap  fdi=%%x  mhdr=%%p  *mhdrpp=%%p  xi=%%p(%%x %%p) f_unf=%%p\\n",
+        fdi, mhdr, (mhdrpp ? *mhdrpp : 0), xi, (xi? xi->size: 0), (xi? xi->buf: 0), f_unf);
+
+    Addr rv = 0;
+    Mach_segment_command *sc = (Mach_segment_command *)(1+ mhdr);
+    Addr const reloc = xfind_pages(mhdr, sc, mhdr->ncmds, 0);
+    DPRINTF("do_xmap reloc=%%p\\n", reloc);
     unsigned j;
-
-    DPRINTF((STR_do_xmap(),
-        fdi, mhdr, xi, (xi? xi->size: 0), (xi? xi->buf: 0), f_unf));
-
     for ( j=0; j < mhdr->ncmds; ++j,
-        (sc = (Mach_segment_command const *)(sc->cmdsize + (void const *)sc))
-    ) if (LC_SEGMENT==sc->cmd && sc->vmsize!=0) {
-        Extent xo;
-        size_t mlen = xo.size = sc->filesize;
-        unsigned char  *addr = xo.buf  =        (unsigned char *)sc->vmaddr;
-        unsigned char *haddr =           sc->vmsize +                  addr;
-        size_t frag = (int)addr &~ PAGE_MASK;
-        addr -= frag;
-        mlen += frag;
+        (sc = (Mach_segment_command *)((sc->cmdsize>>2) + (unsigned *)sc))
+    ) {
+        DPRINTF("  #%%d  cmd=%%x  cmdsize=%%x  vmsize=%%x\\n",
+                j, sc->cmd, sc->cmdsize, sc->vmsize);
+        if (LC_SEGMENT==sc->cmd && !sc->vmsize) {
+            // Typical __DWARF info segment for 'rust'
+            struct b_info h;
+            xread(xi, (unsigned char *)&h, sizeof(h));
+            DPRINTF("    0==.vmsize; skipping %%x\\n", h.sz_cpr);
+            xi->buf += h.sz_cpr;
+        }
+        if (LC_SEGMENT==sc->cmd && sc->vmsize) {
+            Extent xo;
+            size_t mlen = xo.size = sc->filesize;
+                          xo.buf  = (void *)(reloc + sc->vmaddr);
+            Addr  addr = (Addr)xo.buf;
+            Addr haddr = sc->vmsize + addr;
+            size_t frag = addr &~ PAGE_MASK;
+            addr -= frag;
+            mlen += frag;
 
-        if (0!=mlen) {
-            // Decompressor can overrun the destination by 3 bytes.  [x86 only]
-            size_t const mlen3 = mlen + (xi ? 3 : 0);
-            unsigned const prot = VM_PROT_READ | VM_PROT_WRITE;
-            unsigned const flags = MAP_FIXED | MAP_PRIVATE |
-                        ((xi || 0==sc->filesize) ? MAP_ANON : 0);
-            int const fdm = ((0==sc->filesize) ? MAP_ANON_FD : fdi);
-            off_t_upx_stub const offset = sc->fileoff + fat_offset;
+            DPRINTF("    mlen=%%p  frag=%%p  addr=%%p\\n", mlen, frag, addr);
+            if (0!=mlen) { // In particular, omitted for __PAGEZERO
+                size_t const mlen3 = mlen
+    #if defined(__x86_64__)  //{
+                    // Decompressor can overrun the destination by 3 bytes.  [x86 only]
+                    + (xi ? 3 : 0)
+    #endif  //}
+                    ;
+                unsigned const prot = VM_PROT_READ | VM_PROT_WRITE;
+                // MAP_FIXED: xfind_pages() reserved them, so use them!
+                unsigned const flags = MAP_FIXED | MAP_PRIVATE |
+                                ((xi || 0==sc->filesize) ? MAP_ANON    : 0);
+                int const fdm = ((xi || 0==sc->filesize) ? MAP_ANON_FD : fdi);
+                off_t_upx_stub const offset = sc->fileoff + fat_offset;
 
-            DPRINTF((STR_mmap(), addr, mlen3, prot, flags, fdm, offset));
-            if (addr !=     mmap(addr, mlen3, prot, flags, fdm, offset, 0)) {
-                err_exit(8);
+                DPRINTF("mmap  addr=%%p  len=%%p  prot=%%x  flags=%%x  fd=%%d  off=%%p  reloc=%%p\\n",
+                    addr, mlen3, prot, flags, fdm, offset, reloc);
+                {
+                    Addr maddr = (Addr)mmap((void *)addr, mlen3, prot, flags, fdm, offset, 0);
+                    DPRINTF("maddr=%%p\\n", maddr);
+                    if (maddr != addr) {
+                        err_exit(8);
+                    }
+                    addr = maddr;
+                }
+                if (mhdrpp && !*mhdrpp) { // MH_DYLINKER
+                    *mhdrpp = (Mach_header*)addr;
+                }
+            }
+            if (xi && 0!=sc->filesize) {
+                if (0==sc->fileoff /*&& 0!=mhdrpp*/) {
+                    *mhdrpp = (Mach_header *)(void *)addr;
+                }
+                unpackExtent(xi, &xo, f_exp, f_unf);
+            }
+            DPRINTF("xi=%%p  mlen=%%p  fileoff=%%p  nsects=%%d\\n",
+                xi, mlen, sc->fileoff, sc->nsects);
+            if (xi && mlen && !sc->fileoff && sc->nsects) {
+                // main target __TEXT segment at beginning of file with sections (__text)
+                // Use upto 2 words of header padding for the escape hatch.
+                // fold.S could do this easier, except PROT_WRITE is missing then.
+                union {
+                    unsigned char  *p0;
+                    unsigned short *p1;
+                    unsigned int   *p2;
+                    unsigned long  *p3;
+                } u;
+                u.p0 = (unsigned char *)addr;
+                Mach_segment_command *segp = (Mach_segment_command *)((((char *)sc - (char *)mhdr)>>2) + u.p2);
+                Mach_section_command *const secp = (Mach_section_command *)(1+ segp);
+                unsigned *hatch= -2+ (secp->offset>>2) + u.p2;
+                DPRINTF("hatch=%%p  segp=%%p  secp=%%p  secp->offset=%%p  mhdr=%%p\\n", hatch, segp, secp, secp->offset, addr);
+    #if defined(__aarch64__)  //{
+                hatch[0] = 0xd4000001;  // svc #0  // syscall
+                hatch[1] = 0xd65f03c0;  // ret
+    #elif defined(__arm__)  //}{
+                hatch[0] = 0xef000000;  // svc 0x0  // syscall
+                hatch[1] = 0xe12fff1e;  // bx lr
+    #elif defined(__x86_64__)  //}{
+                hatch[0] = 0xc3050f90;  // nop; syscall; ret
+    #elif defined(__i386__)  //}{
+                hatch[0] = 0xc3050f90;  // nop; syscall; ret
+    #endif  //}
+                rv = (Addr)hatch;
+            }
+            /*bzero(addr, frag);*/  // fragment at lo end
+            frag = (-mlen) &~ PAGE_MASK;  // distance to next page boundary
+            bzero((void *)(mlen+addr), frag);  // fragment at hi end
+            if (0!=mlen && 0!=mprotect((void *)addr, mlen, sc->initprot)) {
+                err_exit(10);
+    ERR_LAB
+            }
+            addr += mlen + frag;  /* page boundary on hi end */
+            if (
+    #if SIMULATE_ON_LINUX_EABI4  /*{*/
+                0!=addr &&
+    #endif  /*}*/
+                            addr < haddr) { // need pages for .bss
+                if (0!=addr && addr != (Addr)mmap((void *)addr, haddr - addr, sc->initprot,
+                        MAP_FIXED | MAP_PRIVATE | MAP_ANON, MAP_ANON_FD, 0, 0) ) {
+                    err_exit(9);
+                }
+            }
+            else if (xi) { // cleanup if decompressor overrun crosses page boundary
+                mlen = ~PAGE_MASK & (3+ mlen);
+                if (mlen<=3) { // page fragment was overrun buffer only
+                    DPRINTF("munmap  %%x  %%x\\n", addr, mlen);
+                    munmap((char *)addr, mlen);
+                }
             }
         }
-        if (xi && 0!=sc->filesize) {
-            if (0==sc->fileoff && 0!=mhdrpp) {
-                *mhdrpp = (Mach_header *)addr;
-            }
-            unpackExtent(xi, &xo, f_decompress, f_unf);
-        }
-        /*bzero(addr, frag);*/  // fragment at lo end
-        frag = (-mlen) &~ PAGE_MASK;  // distance to next page boundary
-        bzero(mlen+addr, frag);  // fragment at hi end
-        if (0!=mlen && 0!=mprotect(addr, mlen, sc->initprot)) {
-            err_exit(10);
-ERR_LAB
-        }
-        addr += mlen + frag;  /* page boundary on hi end */
-        if (addr < haddr) { // need pages for .bss
-            if (0!=addr && addr != mmap(addr, haddr - addr, sc->initprot,
-                    MAP_FIXED | MAP_PRIVATE | MAP_ANON, MAP_ANON_FD, 0, 0 ) ) {
-                err_exit(9);
-            }
-        }
-        else if (xi) { // cleanup if decompressor overrun crosses page boundary
-            mlen = ~PAGE_MASK & (3+ mlen);
-            if (mlen<=3) { // page fragment was overrun buffer only
-                munmap(addr, mlen);
+        else if (!xi  // dyld
+        && (LC_UNIXTHREAD==sc->cmd || LC_THREAD==sc->cmd)) {
+            Mach_thread_command *const thrc = (Mach_thread_command *)sc;
+            DPRINTF("thread_command= %%p\\n", sc);
+            if (1
+            // FIXME  THREAD_STATE      ==thrc->flavor
+            //    &&  THREAD_STATE_COUNT==thrc->count
+            ) {
+                DPRINTF("thread_state= %%p  flavor=%%d  count=%%x  reloc=%%p\\n",
+                    &thrc->state, thrc->flavor, thrc->count, reloc);
+    #if defined(__aarch64__)  //{
+                rv = reloc + thrc->state.pc;
+    #elif defined(__arm__)  //}{
+                rv = reloc + thrc->state.pc;
+    #elif defined(__x86_64__)  //}{
+                rv = reloc + thrc->state.rip;
+    #elif defined(__i386__)  //}{
+                rv = reloc + thrc->state.eip;
+    #else  //}{
+        #error do_xmap rv $ARCH
+    #endif  //}
             }
         }
     }
-    else if (LC_UNIXTHREAD==sc->cmd || LC_THREAD==sc->cmd) {
-        Mach_thread_command const *const thrc = (Mach_thread_command const *)sc;
-        if (i386_THREAD_STATE      ==thrc->flavor
-        &&  i386_THREAD_STATE_COUNT==thrc->count ) {
-            entry = &thrc->state;
-        }
-    }
-    return entry;
+    DPRINTF("do_xmap= %%p\\n", rv);
+    return rv;
 }
 
-
 /*************************************************************************
 // upx_main - called by our entry code
 //
 **************************************************************************/
 
-DEBUG_STRCON(STR_upx_main,
-    "upx_main szc=%%x  f_dec=%%p  f_unf=%%p  "
-    "  xo=%%p(%%x %%p)  xi=%%p(%%x %%p)  mhdrpp=%%p\\n")
-
-Mach_i386_thread_state const *
+Addr
 upx_main(
     Mach_header **const mhdrpp,  // Out: *mhdrpp= &real Mach_header
     f_unfilter *const f_unf,
@@ -571,7 +708,7 @@ upx_main(
     struct l_info const *const li
 )
 {
-    Mach_i386_thread_state const *entry;
+    Addr entry;
     off_t_upx_stub fat_offset = 0;
     Extent xi, xo, xi0;
     xi.buf  = CONST_CAST(unsigned char *, 1+ (struct p_info const *)(1+ li));  // &b_info
@@ -580,9 +717,11 @@ upx_main(
     xo.size = ((struct b_info const *)(void const *)xi.buf)->sz_unc;
     xi0 = xi;
 
-    DPRINTF((STR_upx_main(),
-        sz_compressed, f_decompress, f_unf, &xo, xo.size, xo.buf,
-        &xi, xi.size, xi.buf, mhdrpp));
+    DPRINTF("upx_main szc=%%x  f_dec=%%p  f_unf=%%p  "
+    "  xo=%%p(%%x %%p)  xi=%%p(%%x %%p)  mhdrpp=%%p\\n",
+        sz_compressed, f_decompress, f_unf,
+        &xo, xo.size, xo.buf,
+        &xi, xi.size, xi.buf, mhdrpp);
 
     // Uncompress Macho headers
     unpackExtent(&xi, &xo, f_decompress, 0);  // never filtered?
